@@ -18,14 +18,13 @@
 #include <afs/param.h>
 #include <afs/stds.h>
 
-#ifndef DJGPP
 #include <windows.h>
 #include <winsock2.h>
-#endif /* !DJGPP */
 #include <string.h>
 #include <stdlib.h>
 #include <osi.h>
 #include "afsd.h"
+#include <WINNT/afsreg.h>
 
 osi_rwlock_t cm_dnlcLock;
 
@@ -106,34 +105,35 @@ InsertEntry(cm_nc_t *tnc)
 
 void 
 cm_dnlcEnter ( cm_scache_t *adp,
-               char        *aname,
+               normchar_t  *nname,
                cm_scache_t *avc )
 {
     cm_nc_t *tnc;
     unsigned int key, skey, new=0;
-    char *ts = aname;
+    normchar_t *ts = nname;
     int safety;
+    int writeLocked = 0;
 
     if (!cm_useDnlc)
 	return ;
 
-    if (!strcmp(aname,".") || !strcmp(aname,".."))
+    if (!cm_NormStrCmp(nname,_C(".")) || !cm_NormStrCmp(nname,_C("..")))
 	return ;
 
     if ( cm_debugDnlc ) 
-	osi_Log3(afsd_logp,"cm_dnlcEnter dir %x name %s scache %x", 
-	    adp, osi_LogSaveString(afsd_logp,aname), avc);
+	osi_Log3(afsd_logp,"cm_dnlcEnter dir %x name %S scache %x", 
+	    adp, osi_LogSaveStringW(afsd_logp,nname), avc);
 
     dnlcHash( ts, key );  /* leaves ts pointing at the NULL */
-    if (ts - aname >= CM_AFSNCNAMESIZE) 
+    if (ts - nname >= CM_AFSNCNAMESIZE) 
 	return ;
     skey = key & (NHSIZE -1);
 
-    lock_ObtainWrite(&cm_dnlcLock);
-    dnlcstats.enters++;
-  
+    InterlockedIncrement(&dnlcstats.enters);
+    lock_ObtainRead(&cm_dnlcLock);
+  retry:
     for (tnc = cm_data.nameHash[skey], safety=0; tnc; tnc = tnc->next, safety++ )
-	if ((tnc->dirp == adp) && (!strcmp(tnc->name, aname)))
+	if ((tnc->dirp == adp) && (!cm_NormStrCmp(tnc->name, nname)))
 	    break;				/* preexisting entry */
 	else if ( tnc->next == cm_data.nameHash[skey])	/* end of list */
 	{
@@ -142,8 +142,11 @@ cm_dnlcEnter ( cm_scache_t *adp,
 	}
 	else if (safety > NCSIZE) 
 	{
-	    dnlcstats.cycles++;
-	    lock_ReleaseWrite(&cm_dnlcLock);
+	    InterlockedIncrement(&dnlcstats.cycles);
+            if (writeLocked)
+                lock_ReleaseWrite(&cm_dnlcLock);
+            else
+                lock_ReleaseRead(&cm_dnlcLock);
 
 	    if ( cm_debugDnlc )
                 osi_Log0(afsd_logp, "DNLC cycle");
@@ -153,6 +156,11 @@ cm_dnlcEnter ( cm_scache_t *adp,
 	
     if ( !tnc )
     {
+        if ( !writeLocked ) {
+            lock_ConvertRToW(&cm_dnlcLock);
+            writeLocked = 1;
+            goto retry;
+        }
 	new = 1;	/* entry does not exist, we are creating a new entry*/
 	tnc = GetMeAnEntry();
     }
@@ -161,13 +169,16 @@ cm_dnlcEnter ( cm_scache_t *adp,
 	tnc->dirp = adp;
 	tnc->vp = avc;
 	tnc->key = key;
-	memcpy (tnc->name, aname, ts-aname+1); /* include the NULL */
+	memcpy (tnc->name, nname, (ts-nname+1)*sizeof(normchar_t)); /* include the NULL */
 
     	if ( new )	/* insert entry only if it is newly created */ 
-		InsertEntry(tnc);
+            InsertEntry(tnc);
 
     }
-    lock_ReleaseWrite(&cm_dnlcLock);
+    if (writeLocked)
+        lock_ReleaseWrite(&cm_dnlcLock);
+    else
+        lock_ReleaseRead(&cm_dnlcLock);
 
     if ( !tnc)
 	cm_dnlcPurge();
@@ -181,25 +192,30 @@ cm_dnlcLookup (cm_scache_t *adp, cm_lookupSearch_t* sp)
 {
     cm_scache_t * tvc;
     unsigned int key, skey;
-    char* aname = sp->searchNamep;
-    char *ts = aname;
+    normchar_t* nname = sp->nsearchNamep;
+    normchar_t *ts = nname;
     cm_nc_t * tnc, * tnc_begin;
     int safety, match;
   
     if (!cm_useDnlc)
 	return NULL;
+
     if ( cm_debugDnlc ) 
-	osi_Log2(afsd_logp, "cm_dnlcLookup dir %x name %s", 
-		adp, osi_LogSaveString(afsd_logp,aname));
+	osi_Log2(afsd_logp, "cm_dnlcLookup dir %x name %S", 
+		adp, osi_LogSaveStringW(afsd_logp,nname));
 
     dnlcHash( ts, key );  /* leaves ts pointing at the NULL */
-    if (ts - aname >= CM_AFSNCNAMESIZE) 
+
+    if (ts - nname >= CM_AFSNCNAMESIZE) {
+        InterlockedIncrement(&dnlcstats.lookups);
+        InterlockedIncrement(&dnlcstats.misses);
 	return NULL;
+    }
 
     skey = key & (NHSIZE -1);
 
     lock_ObtainRead(&cm_dnlcLock);
-    dnlcstats.lookups++;	     /* Is a dnlcread lock sufficient? */
+    InterlockedIncrement(&dnlcstats.lookups);
 
     ts = 0;
     tnc_begin = cm_data.nameHash[skey];
@@ -209,26 +225,26 @@ cm_dnlcLookup (cm_scache_t *adp, cm_lookupSearch_t* sp)
 	if (tnc->dirp == adp) 
 	{
         if( cm_debugDnlc ) 
-            osi_Log1(afsd_logp,"Looking at [%s]",
-                     osi_LogSaveString(afsd_logp,tnc->name));
+            osi_Log1(afsd_logp,"Looking at [%S]",
+                     osi_LogSaveStringW(afsd_logp,tnc->name));
 
 	    if ( sp->caseFold ) 	/* case insensitive */
 	    {
-            match = cm_stricmp(tnc->name, aname);
+            match = cm_NormStrCmpI(tnc->name, nname);
             if ( !match )	/* something matches */
             {
                 tvc = tnc->vp;
                 ts = tnc->name;
 
                 /* determine what type of match it is */
-                if ( !strcmp(tnc->name, aname))
+                if ( !cm_NormStrCmp(tnc->name, nname))
                 {	
                     /* exact match. */
                     sp->ExactFound = 1;
 
                     if( cm_debugDnlc )
-                        osi_Log1(afsd_logp,"DNLC found exact match [%s]",
-                                 osi_LogSaveString(afsd_logp,tnc->name));
+                        osi_Log1(afsd_logp,"DNLC found exact match [%S]",
+                                 osi_LogSaveStringW(afsd_logp,tnc->name));
                     break;
                 }
                 else if ( cm_NoneUpper(tnc->name))
@@ -242,7 +258,7 @@ cm_dnlcLookup (cm_scache_t *adp, cm_lookupSearch_t* sp)
 	    }
 	    else			/* case sensitive */
 	    {
-            match = strcmp(tnc->name, aname);
+            match = cm_NormStrCmp(tnc->name, nname);
             if ( !match ) /* found a match */
             {
                 sp->ExactFound = 1;
@@ -258,7 +274,7 @@ cm_dnlcLookup (cm_scache_t *adp, cm_lookupSearch_t* sp)
 	}
 	else if (tnc->next == tnc_begin || safety > NCSIZE) 
 	{
-	    dnlcstats.cycles++;
+	    InterlockedIncrement(&dnlcstats.cycles);
 	    lock_ReleaseRead(&cm_dnlcLock);
 
 	    if ( cm_debugDnlc ) 
@@ -269,14 +285,14 @@ cm_dnlcLookup (cm_scache_t *adp, cm_lookupSearch_t* sp)
     }
 
     if(cm_debugDnlc && ts) {
-        osi_Log3(afsd_logp, "DNLC matched [%s] for [%s] with vnode[%ld]",
-                 osi_LogSaveString(afsd_logp,ts),
-                 osi_LogSaveString(afsd_logp,aname),
+        osi_Log3(afsd_logp, "DNLC matched [%S] for [%W] with vnode[%ld]",
+                 osi_LogSaveStringW(afsd_logp,ts),
+                 osi_LogSaveStringW(afsd_logp,nname),
                  (long) tvc->fid.vnode);
     }
 
     if (!tvc) 
-        dnlcstats.misses++; 	/* Is a dnlcread lock sufficient? */
+        InterlockedIncrement(&dnlcstats.misses);
     else 
     {
         sp->found = 1;
@@ -321,32 +337,32 @@ RemoveEntry (cm_nc_t *tnc, unsigned int key)
 
 
 void 
-cm_dnlcRemove (cm_scache_t *adp, char *aname)
+cm_dnlcRemove (cm_scache_t *adp, normchar_t *nname)
 {
     unsigned int key, skey, error=0;
     int found= 0, safety;
-    char *ts = aname;
+    normchar_t *ts = nname;
     cm_nc_t *tnc, *tmp;
   
     if (!cm_useDnlc)
 	return ;
 
     if ( cm_debugDnlc )
-	osi_Log2(afsd_logp, "cm_dnlcRemove dir %x name %s", 
-		adp, osi_LogSaveString(afsd_logp,aname));
+	osi_Log2(afsd_logp, "cm_dnlcRemove dir %x name %S", 
+		adp, osi_LogSaveStringW(afsd_logp,nname));
 
     dnlcHash( ts, key );  /* leaves ts pointing at the NULL */
-    if (ts - aname >= CM_AFSNCNAMESIZE) 
+    if (ts - nname >= CM_AFSNCNAMESIZE) 
 	return ;
 
     skey = key & (NHSIZE -1);
     lock_ObtainWrite(&cm_dnlcLock);
-    dnlcstats.removes++;
+    InterlockedIncrement(&dnlcstats.removes);
 
     for (tnc = cm_data.nameHash[skey], safety=0; tnc; safety++) 
     {
 	if ( (tnc->dirp == adp) && (tnc->key == key) 
-			&& !strcmp(tnc->name,aname) )
+             && !cm_NormStrCmp(tnc->name,nname) )
 	{
 	    tmp = tnc->next;
     	    error = RemoveEntry(tnc, skey);
@@ -365,7 +381,7 @@ cm_dnlcRemove (cm_scache_t *adp, char *aname)
 	    tnc = tnc->next;
 	if ( safety > NCSIZE )
 	{
-	    dnlcstats.cycles++;
+	    InterlockedIncrement(&dnlcstats.cycles);
 	    lock_ReleaseWrite(&cm_dnlcLock);
 
 	    if ( cm_debugDnlc ) 
@@ -397,7 +413,7 @@ cm_dnlcPurgedp (cm_scache_t *adp)
 	osi_Log1(afsd_logp, "cm_dnlcPurgedp dir %x", adp);
 
     lock_ObtainWrite(&cm_dnlcLock);
-    dnlcstats.purgeds++;
+    InterlockedIncrement(&dnlcstats.purgeds);
 
     for (i=0; i<NCSIZE && !err; i++) 
     {
@@ -434,7 +450,7 @@ cm_dnlcPurgevp (cm_scache_t *avc)
 	osi_Log1(afsd_logp, "cm_dnlcPurgevp scache %x", avc);
 
     lock_ObtainWrite(&cm_dnlcLock);
-    dnlcstats.purgevs++;
+    InterlockedIncrement(&dnlcstats.purgevs);
 
     for (i=0; i<NCSIZE && !err ; i++) 
     {
@@ -469,7 +485,7 @@ void cm_dnlcPurge(void)
 	osi_Log0(afsd_logp, "cm_dnlcPurge");
 
     lock_ObtainWrite(&cm_dnlcLock);
-    dnlcstats.purges++;
+    InterlockedIncrement(&dnlcstats.purges);
     
     cm_data.ncfreelist = (cm_nc_t *) 0;
     memset (cm_data.nameCache, 0, sizeof(cm_nc_t) * NCSIZE);
@@ -493,7 +509,7 @@ cm_dnlcPurgeVol(AFSFid *fidp)
     if (!cm_useDnlc)
         return ;
 
-    dnlcstats.purgevols++;
+    InterlockedIncrement(&dnlcstats.purgevols);
     cm_dnlcPurge();
 }
 
@@ -627,9 +643,29 @@ void
 cm_dnlcInit(int newFile)
 {
     int i;
+    HKEY parmKey;
+    DWORD dummyLen;
+    DWORD dwValue;
+    DWORD code;
 
-    if (!cm_useDnlc)
-        return ;
+    code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSREG_CLT_OPENAFS_SUBKEY,
+                         0, KEY_QUERY_VALUE, &parmKey);
+    if (code == ERROR_SUCCESS) {
+        dummyLen = sizeof(DWORD);
+        code = RegQueryValueEx(parmKey, "UseDNLC", NULL, NULL,
+                                (BYTE *) &dwValue, &dummyLen);
+        if (code == ERROR_SUCCESS)
+            cm_useDnlc = dwValue ? 1 : 0;
+        afsi_log("CM UseDNLC = %d", cm_useDnlc);
+
+        dummyLen = sizeof(DWORD);
+        code = RegQueryValueEx(parmKey, "DebugDNLC", NULL, NULL,
+                                (BYTE *) &dwValue, &dummyLen);
+        if (code == ERROR_SUCCESS)
+            cm_debugDnlc = dwValue ? 1 : 0;
+        afsi_log("CM DebugDNLC = %d", cm_debugDnlc);
+        RegCloseKey (parmKey);
+    }
 
     if ( cm_debugDnlc )
 	osi_Log0(afsd_logp,"cm_dnlcInit");
