@@ -1,13 +1,9 @@
 #include <afs/param.h>
 #include <afs/stds.h>
 
-#ifndef DJGPP
 #include <windows.h>
 #include <winreg.h>
 #include <winsock2.h>
-#else
-#include <netdb.h>
-#endif /* !DJGPP */
 #include <stdlib.h>
 #include <malloc.h>
 #include <string.h>
@@ -34,14 +30,11 @@ int cm_freelanceEnabled = 1;
 time_t FakeFreelanceModTime = 0x3b49f6e2;
 
 static int freelance_ShutdownFlag = 0;
-#if !defined(DJGPP)
 static HANDLE hFreelanceChangeEvent = 0;
 static HANDLE hFreelanceSymlinkChangeEvent = 0;
-#endif
 
 void cm_InitFakeRootDir();
 
-#if !defined(DJGPP)
 void cm_FreelanceChangeNotifier(void * parmp) {
     HKEY   hkFreelance = 0;
 
@@ -131,25 +124,20 @@ void cm_FreelanceSymlinkChangeNotifier(void * parmp) {
         }
     }
 }
-#endif
 
 void                                          
 cm_FreelanceShutdown(void)                    
 {                                             
     freelance_ShutdownFlag = 1;               
-#if !defined(DJGPP)                           
     if (hFreelanceChangeEvent != 0)           
         thrd_SetEvent(hFreelanceChangeEvent); 
     if (hFreelanceSymlinkChangeEvent != 0)           
         thrd_SetEvent(hFreelanceSymlinkChangeEvent); 
-#endif                                        
 }                                             
 
 void cm_InitFreelance() {
-#if !defined(DJGPP)
     thread_t phandle;
     int lpid;
-#endif
 
     lock_InitializeMutex(&cm_Freelance_Lock, "Freelance Lock");
 
@@ -162,7 +150,6 @@ void cm_InitFreelance() {
     cm_InitFakeRootDir();
     // --- end of yj code
 
-#if !defined(DJGPP)
     /* Start the registry monitor */
     phandle = thrd_Create(NULL, 65536, (ThreadFunc) cm_FreelanceChangeNotifier,
                           NULL, 0, &lpid, "cm_FreelanceChangeNotifier");
@@ -173,7 +160,6 @@ void cm_InitFreelance() {
                           NULL, 0, &lpid, "cm_FreelanceSymlinkChangeNotifier");
     osi_assertx(phandle != NULL, "cm_FreelanceSymlinkChangeNotifier thread create failure");
     thrd_CloseHandle(phandle);
-#endif
 }
 
 /* yj: Initialization of the fake root directory */
@@ -419,6 +405,7 @@ int cm_reInitLocalMountPoints() {
                      lscpp = &tscp->nextp, tscp = tscp->nextp) {
                     if (tscp == scp) {
                         *lscpp = scp->nextp;
+                        scp->nextp = NULL;
 			lock_ObtainWrite(&scp->rw);
                         scp->flags &= ~CM_SCACHEFLAG_INHASH;
 			lock_ReleaseWrite(&scp->rw);
@@ -468,16 +455,13 @@ long cm_InitLocalMountPoints() {
     char hdir[260];
     long code;
     char rootCellName[256];
-#if !defined(DJGPP)
     HKEY hkFreelance = 0, hkFreelanceSymlinks = 0;
     DWORD dwType, dwSize;
     DWORD dwMountPoints = 0;
     DWORD dwIndex;
     DWORD dwSymlinks = 0;
     FILETIME ftLastWriteTime;
-#endif
 
-#if !defined(DJGPP)
     if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
                       AFSREG_CLT_OPENAFS_SUBKEY "\\Freelance",
                       0,
@@ -669,7 +653,6 @@ long cm_InitLocalMountPoints() {
         RegCloseKey(hkFreelance);
         return 0;
     }
-#endif
 
     /* What follows is the old code to read freelance mount points 
      * out of a text file modified to copy the data into the registry
@@ -686,7 +669,6 @@ long cm_InitLocalMountPoints() {
         fp = fopen(hdir, "r");
     }
 
-#if !defined(DJGPP)
     RegCreateKeyEx( HKEY_LOCAL_MACHINE, 
                     AFSREG_CLT_OPENAFS_SUBKEY "\\Freelance",
                     0,
@@ -697,12 +679,9 @@ long cm_InitLocalMountPoints() {
                     &hkFreelance,
                     NULL);
     dwIndex = 0;
-#endif
 
     if (!fp) {
-#if !defined(DJGPP)
         RegCloseKey(hkFreelance);
-#endif
         rootCellName[0] = '.';
       	code = cm_GetRootCellName(&rootCellName[1]);
         if (code == 0) {
@@ -759,7 +738,6 @@ long cm_InitLocalMountPoints() {
         if (t2)
             *(t2+1) = '\0';
 
-#if !defined(DJGPP)
         if ( hkFreelance ) {
             char szIndex[16];
             /* we are migrating to the registry */
@@ -768,7 +746,6 @@ long cm_InitLocalMountPoints() {
             dwSize = (DWORD)strlen(line) + 1;
             RegSetValueEx( hkFreelance, szIndex, 0, dwType, line, dwSize);
         }
-#endif 
 
         // line is not empty, so let's parse it
         t = strchr(line, '#');
@@ -795,12 +772,10 @@ long cm_InitLocalMountPoints() {
         aLocalMountPoint++;
     }
     fclose(fp);
-#if !defined(DJGPP)
     if ( hkFreelance ) {
         RegCloseKey(hkFreelance);
         DeleteFile(hdir);
     }
-#endif
     return 0;
 }
 
@@ -808,7 +783,6 @@ int cm_getNoLocalMountPoints() {
     return cm_noLocalMountPoints;
 }
 
-#if !defined(DJGPP)
 long cm_FreelanceMountPointExists(char * filename, int prefix_ok)
 {
     char* cp;
@@ -873,7 +847,7 @@ long cm_FreelanceMountPointExists(char * filename, int prefix_ok)
             memcpy(shortname, line, cp-line);
             shortname[cp-line]=0;
 
-            if (!stricmp(shortname, filename)) {
+            if (!cm_stricmp_utf8(shortname, filename)) {
                 found = 1;
                 break;
             }
@@ -956,7 +930,7 @@ long cm_FreelanceSymlinkExists(char * filename, int prefix_ok)
             memcpy(shortname, line, cp-line);
             shortname[cp-line]=0;
 
-            if (!stricmp(shortname, filename)) {
+            if (!cm_stricmp_utf8(shortname, filename)) {
                 found = 1;
                 break;
             }
@@ -968,7 +942,6 @@ long cm_FreelanceSymlinkExists(char * filename, int prefix_ok)
 
     return found;
 }
-#endif
 
 long cm_FreelanceAddMount(char *filename, char *cellname, char *volume, int rw, cm_fid_t *fidp)
 {
@@ -978,12 +951,10 @@ long cm_FreelanceAddMount(char *filename, char *cellname, char *volume, int rw,
     char fullname[200];
     int n;
     int alias = 0;
-#if !defined(DJGPP)
     HKEY hkFreelance = 0;
     DWORD dwType, dwSize;
     DWORD dwMountPoints;
     DWORD dwIndex;
-#endif
 
     /* before adding, verify the cell name; if it is not a valid cell,
        don't add the mount point.
@@ -1006,18 +977,15 @@ long cm_FreelanceAddMount(char *filename, char *cellname, char *volume, int rw,
             return -1;
     }
 
-#if !defined(DJGPP)
     if ( cm_FreelanceMountPointExists(filename, 0) ||
          cm_FreelanceSymlinkExists(filename, 0) )
         return -1;
-#endif
     
     osi_Log1(afsd_logp,"Freelance Adding Mount for Cell: %s", 
               osi_LogSaveString(afsd_logp,cellname));
 
     lock_ObtainMutex(&cm_Freelance_Lock);
 
-#if !defined(DJGPP)
     if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
                       AFSREG_CLT_OPENAFS_SUBKEY "\\Freelance",
                       0,
@@ -1072,7 +1040,6 @@ long cm_FreelanceAddMount(char *filename, char *cellname, char *volume, int rw,
         }
         RegCloseKey(hkFreelance);
     } else 
-#endif
     {
         cm_GetConfigDir(hfile, sizeof(hfile));
         strcat(hfile, AFS_FREELANCE_INI);
@@ -1109,16 +1076,13 @@ long cm_FreelanceRemoveMount(char *toremove)
     char hfile[260], hfile2[260];
     FILE *fp1, *fp2;
     int found=0;
-#if !defined(DJGPP)
     HKEY hkFreelance = 0;
     DWORD dwType, dwSize;
     DWORD dwMountPoints;
     DWORD dwIndex;
-#endif
 
     lock_ObtainMutex(&cm_Freelance_Lock);
 
-#if !defined(DJGPP)
     if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
                       AFSREG_CLT_OPENAFS_SUBKEY "\\Freelance",
                       0,
@@ -1160,7 +1124,6 @@ long cm_FreelanceRemoveMount(char *toremove)
         }
         RegCloseKey(hkFreelance);
     } else 
-#endif
     {
         cm_GetConfigDir(hfile, sizeof(hfile));
         strcat(hfile, AFS_FREELANCE_INI);
@@ -1216,12 +1179,10 @@ long cm_FreelanceAddSymlink(char *filename, char *destination, cm_fid_t *fidp)
     char line[512];
     char fullname[200];
     int alias = 0;
-#if !defined(DJGPP)
     HKEY hkFreelanceSymlinks = 0;
     DWORD dwType, dwSize;
     DWORD dwSymlinks;
     DWORD dwIndex;
-#endif
 
     /* before adding, verify the filename.  If it is already in use, either as 
      * as mount point or a cellname, do not permit the creation of the symlink.
@@ -1236,23 +1197,20 @@ long cm_FreelanceAddSymlink(char *filename, char *destination, cm_fid_t *fidp)
     fullname[0] = '\0';
     if (filename[0] == '.') {
         cm_GetCell_Gen(&filename[1], fullname, CM_FLAG_CREATE);
-        if (stricmp(&filename[1],fullname) == 0)
+        if (cm_stricmp_utf8(&filename[1],fullname) == 0)
             return CM_ERROR_EXISTS;
     } else {
         cm_GetCell_Gen(filename, fullname, CM_FLAG_CREATE);
-        if (stricmp(filename,fullname) == 0)
+        if (cm_stricmp_utf8(filename,fullname) == 0)
             return CM_ERROR_EXISTS;
     }
 
-#if !defined(DJGPP)
     if ( cm_FreelanceMountPointExists(filename, 0) ||
          cm_FreelanceSymlinkExists(filename, 0) )
         return CM_ERROR_EXISTS;
-#endif
 
     lock_ObtainMutex(&cm_Freelance_Lock);
 
-#if !defined(DJGPP)
     if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, 
                         AFSREG_CLT_OPENAFS_SUBKEY "\\Freelance\\Symlinks",
                         0,
@@ -1307,7 +1265,6 @@ long cm_FreelanceAddSymlink(char *filename, char *destination, cm_fid_t *fidp)
         }
         RegCloseKey(hkFreelanceSymlinks);
     } 
-#endif
     lock_ReleaseMutex(&cm_Freelance_Lock);
 
     /* cm_reInitLocalMountPoints(); */
@@ -1323,16 +1280,13 @@ long cm_FreelanceRemoveSymlink(char *toremove)
     char line[512];
     char shortname[200];
     int found=0;
-#if !defined(DJGPP)
     HKEY hkFreelanceSymlinks = 0;
     DWORD dwType, dwSize;
     DWORD dwSymlinks;
     DWORD dwIndex;
-#endif
 
     lock_ObtainMutex(&cm_Freelance_Lock);
 
-#if !defined(DJGPP)
     if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
                       AFSREG_CLT_OPENAFS_SUBKEY "\\Freelance\\Symlinks",
                       0,
@@ -1372,7 +1326,6 @@ long cm_FreelanceRemoveSymlink(char *toremove)
         }
         RegCloseKey(hkFreelanceSymlinks);
     }
-#endif
     
     lock_ReleaseMutex(&cm_Freelance_Lock);
     if (found) {
