@@ -10,13 +10,12 @@
 #include <afs/param.h>
 #include <afs/stds.h>
 
-#ifndef DJGPP
 #include <windows.h>
-#endif /* !DJGPP */
 #include <malloc.h>
 #include <string.h>
 
 #include "afsd.h"
+#include "smb.h"
 #include <osi.h>
 #include <rx/rx.h>
 
@@ -106,7 +105,7 @@ void cm_ReleaseUser(cm_user_t *userp)
         return;
 
     lock_ObtainWrite(&cm_userLock);
-    osi_assert(userp->refCount-- > 0);
+    osi_assertx(userp->refCount-- > 0, "cm_user_t refCount 0");
     if (userp->refCount == 0) {
         lock_FinalizeMutex(&userp->mx);
         for (ucp = userp->cellInfop; ucp; ucp = ncp) {
@@ -137,7 +136,7 @@ void cm_HoldUserVCRef(cm_user_t *userp)
 void cm_ReleaseUserVCRef(cm_user_t *userp)
 {
     lock_ObtainMutex(&userp->mx);
-    osi_assert(userp->vcRefs-- > 0);
+    osi_assertx(userp->vcRefs-- > 0, "cm_user_t refCount 0");
     lock_ReleaseMutex(&userp->mx);
 }       
 
@@ -162,7 +161,7 @@ void cm_CheckTokenCache(time_t now)
     /* 
      * For every vcp, get the user and check his tokens 
      */
-    lock_ObtainWrite(&smb_rctLock);
+    lock_ObtainRead(&smb_rctLock);
     for (vcp=smb_allVCsp; vcp; vcp=vcp->nextp) {
         for (usersp=vcp->usersp; usersp; usersp=usersp->nextp) {
             if (usersp->unp) {
@@ -194,5 +193,22 @@ void cm_CheckTokenCache(time_t now)
             }
         }
     }
-    lock_ReleaseWrite(&smb_rctLock);
+    lock_ReleaseRead(&smb_rctLock);
 }
+
+#ifdef USE_ROOT_TOKENS
+/*
+ * Service/Parameters/RootTokens/<cellname>/
+ * -> UseLSA
+ * -> Keytab (required if UseLSA is 0)
+ * -> Principal (required if there is more than one principal in the keytab)
+ * -> Realm (required if realm is not upper-case of <cellname>
+ * -> RequireEncryption 
+ */
+
+void
+cm_RefreshRootTokens(void)
+{
+
+}
+#endif 
