@@ -10,13 +10,9 @@
 #include <afs/param.h>
 #include <afs/stds.h>
 
-#ifndef DJGPP
 #include <windows.h>
 #include <winsock2.h>
 #include <nb30.h>
-#else
-#include <sys/socket.h>
-#endif /* !DJGPP */
 #include <string.h>
 #include <malloc.h>
 #include "afsd.h"
@@ -106,6 +102,7 @@ void cm_InitVolume(int newFile, long maxVols)
 
                 lock_InitializeRWLock(&volp->rw, "cm_volume_t rwlock");
                 volp->flags |= CM_VOLUMEFLAG_RESET;
+                volp->flags &= ~CM_VOLUMEFLAG_UPDATING_VL;
                 for (volType = RWVOL; volType < NUM_VOL_TYPES; volType++) {
                     volp->vol[volType].state = vl_unknown;
                     volp->vol[volType].serversp = NULL;
@@ -188,12 +185,6 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
 #endif
     afs_uint32 volType;
 
-    /* clear out old bindings */
-    for ( volType = RWVOL; volType < NUM_VOL_TYPES; volType++) {
-        if (volp->vol[volType].serversp)
-            cm_FreeServerList(&volp->vol[volType].serversp, CM_FREESERVERLIST_DELETE);
-    }
-
 #ifdef AFS_FREELANCE_CLIENT
     if ( cellp->cellID == AFS_FAKE_ROOT_CELL_ID && volp->vol[RWVOL].ID == AFS_FAKE_ROOT_VOL_ID ) 
     {
@@ -207,11 +198,24 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
 #endif
     {
         while (volp->flags & CM_VOLUMEFLAG_UPDATING_VL) {
+            osi_Log3(afsd_logp, "cm_UpdateVolumeLocation sleeping name %s:%s flags 0x%x", 
+                     volp->cellp->name, volp->namep, volp->flags);
             osi_SleepW((LONG_PTR) &volp->flags, &volp->rw);
             lock_ObtainWrite(&volp->rw);
-
-            if (!(volp->flags & CM_VOLUMEFLAG_RESET))
+            osi_Log3(afsd_logp, "cm_UpdateVolumeLocation awake name %s:%s flags 0x%x", 
+                     volp->cellp->name, volp->namep, volp->flags);
+            if (!(volp->flags & CM_VOLUMEFLAG_RESET)) {
+                osi_Log3(afsd_logp, "cm_UpdateVolumeLocation nothing to do, waking others name %s:%s flags 0x%x", 
+                         volp->cellp->name, volp->namep, volp->flags);
+                osi_Wakeup((LONG_PTR) &volp->flags);
                 return 0;
+            }
+        }
+
+        /* clear out old bindings */
+        for ( volType = RWVOL; volType < NUM_VOL_TYPES; volType++) {
+            if (volp->vol[volType].serversp)
+                cm_FreeServerList(&volp->vol[volType].serversp, CM_FREESERVERLIST_DELETE);
         }
 
         volp->flags |= CM_VOLUMEFLAG_UPDATING_VL;
@@ -223,22 +227,27 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
         /* now we have volume structure locked and held; make RPC to fill it */
 	osi_Log2(afsd_logp, "CALL VL_GetEntryByName{UNO} name %s:%s", volp->cellp->name, volp->namep);
         do {
+            struct rx_connection * rxconnp;
+
             code = cm_ConnByMServers(cellp->vlServersp, userp, reqp, &connp);
             if (code) 
                 continue;
+
+            rxconnp = cm_GetRxConn(connp);
 #ifdef MULTIHOMED
-            code = VL_GetEntryByNameU(connp->callp, volp->namep, &uvldbEntry);
+            code = VL_GetEntryByNameU(rxconnp, volp->namep, &uvldbEntry);
             method = 2;
             if ( code == RXGEN_OPCODE ) 
 #endif
             {
-                code = VL_GetEntryByNameN(connp->callp, volp->namep, &nvldbEntry);
+                code = VL_GetEntryByNameN(rxconnp, volp->namep, &nvldbEntry);
                 method = 1;
             }
             if ( code == RXGEN_OPCODE ) {
-                code = VL_GetEntryByNameO(connp->callp, volp->namep, &vldbEntry);
+                code = VL_GetEntryByNameO(rxconnp, volp->namep, &vldbEntry);
                 method = 0;
             }
+            rx_PutConnection(rxconnp);
         } while (cm_Analyze(connp, userp, reqp, NULL, NULL, cellp->vlServersp, NULL, code));
         code = cm_MapVLRPCError(code, reqp);
 	if ( code )
@@ -263,22 +272,27 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
 	osi_Log2(afsd_logp, "CALL VL_GetEntryByName{UNO} name %s:%s", volp->cellp->name, 
                  osi_LogSaveString(afsd_logp,name));
         do {
+            struct rx_connection * rxconnp;
+
             code = cm_ConnByMServers(cellp->vlServersp, userp, reqp, &connp);
             if (code) 
                 continue;
+
+            rxconnp = cm_GetRxConn(connp);
 #ifdef MULTIHOMED
-            code = VL_GetEntryByNameU(connp->callp, name, &uvldbEntry);
+            code = VL_GetEntryByNameU(connp->rxconnp, name, &uvldbEntry);
             method = 2;
             if ( code == RXGEN_OPCODE ) 
 #endif
             {
-                code = VL_GetEntryByNameN(connp->callp, name, &nvldbEntry);
+                code = VL_GetEntryByNameN(connp->rxconnp, name, &nvldbEntry);
                 method = 1;
             }
             if ( code == RXGEN_OPCODE ) {
-                code = VL_GetEntryByNameO(connp->callp, name, &vldbEntry);
+                code = VL_GetEntryByNameO(connp->rxconnp, name, &vldbEntry);
                 method = 0;
             }
+            rx_PutConnection(rxconnp);
         } while (cm_Analyze(connp, userp, reqp, NULL, NULL, cellp->vlServersp, NULL, code));
         code = cm_MapVLRPCError(code, reqp);
 	if ( code )
@@ -342,7 +356,7 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
             rwID = uvldbEntry.volumeId[0];
             roID = uvldbEntry.volumeId[1];
             bkID = uvldbEntry.volumeId[2];
-            for ( i=0, j=0; i<nServers && j<NMAXNSERVERS; i++ ) {
+            for ( i=0, j=0; code == 0 && i<nServers && j<NMAXNSERVERS; i++ ) {
                 if ( !(uvldbEntry.serverFlags[i] & VLSERVER_FLAG_UUID) ) {
                     serverFlags[j] = uvldbEntry.serverFlags[i];
                     serverNumber[j] = uvldbEntry.serverNumber[i].time_low;
@@ -360,18 +374,24 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
                     memset((char *)&addrs, 0, sizeof(addrs));
 
                     do {
+                        struct rx_connection *rxconnp;
+
                         code = cm_ConnByMServers(cellp->vlServersp, userp, reqp, &connp);
                         if (code) 
                             continue;
                    
-                        code = VL_GetAddrsU(connp->callp, &attrs, &uuid, &unique, &nentries, &addrs);
-
-                        if (code == 0 && nentries == 0)
-                            code = VL_NOENT;
+                        rxconnp = cm_GetRxConn(connp);
+                        code = VL_GetAddrsU(rxconnp, &attrs, &uuid, &unique, &nentries, &addrs);
+                        rx_PutConnection(rxconnp);
                     } while (cm_Analyze(connp, userp, reqp, NULL, NULL, cellp->vlServersp, NULL, code));
-                    code = cm_MapVLRPCError(code, reqp);
-                    if (code)
-                        return code;
+
+                    if ( code ) {
+                        code = cm_MapVLRPCError(code, reqp);
+                        osi_Log2(afsd_logp, "CALL VL_GetAddrsU serverNumber %u FAILURE, code 0x%x", 
+                                 i, code);
+                        continue;
+                    } 
+                    osi_Log1(afsd_logp, "CALL VL_GetAddrsU serverNumber %u SUCCESS", i);
 
                     addrp = addrs.bulkaddrs_val;
                     for (k = 0; k < nentries && j < NMAXNSERVERS; j++, k++) {
@@ -380,6 +400,9 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
                     }
 
                     free(addrs.bulkaddrs_val);  /* This is wrong */
+
+                    if (nentries == 0)
+                        code = CM_ERROR_INVAL;
                 }
             }
             nServers = j;					/* update the server count */
@@ -458,10 +481,13 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
             tempAddr = htonl(serverNumber[i]);
             tsockAddr.sin_addr.s_addr = tempAddr;
             tsp = cm_FindServer(&tsockAddr, CM_SERVER_FILE);
-            if (!tsp)
-                tsp = cm_NewServer(&tsockAddr, CM_SERVER_FILE,
-                                    cellp, 0);
-
+            if (!tsp) {
+                /* cm_NewServer will probe the server which in turn will
+                 * update the state on the volume group object */
+                lock_ReleaseWrite(&volp->rw);
+                tsp = cm_NewServer(&tsockAddr, CM_SERVER_FILE, cellp, 0);
+                lock_ObtainWrite(&volp->rw);
+            }
             /* if this server was created by fs setserverprefs */
             if ( !tsp->cellp ) 
                 tsp->cellp = cellp;
@@ -571,7 +597,12 @@ long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *
         volp->vol[BACKVOL].state = bkNewstate;
     }
 
+    if (code == 0)
+        volp->flags &= ~CM_VOLUMEFLAG_RESET;
+
     volp->flags &= ~CM_VOLUMEFLAG_UPDATING_VL;
+    osi_Log4(afsd_logp, "cm_UpdateVolumeLocation done, waking others name %s:%s flags 0x%x code 0x%x", 
+             volp->cellp->name, volp->namep, volp->flags, code);
     osi_Wakeup((LONG_PTR) &volp->flags);
 
     return code;
@@ -686,8 +717,6 @@ long cm_FindVolumeByID(cm_cell_t *cellp, afs_uint32 volumeID, cm_user_t *userp,
         code = 0;
         if ((volp->flags & CM_VOLUMEFLAG_RESET) && !(flags & CM_GETVOL_FLAG_NO_RESET)) {
             code = cm_UpdateVolumeLocation(cellp, userp, reqp, volp);
-            if (code == 0)
-                volp->flags &= ~CM_VOLUMEFLAG_RESET;
         }
         lock_ReleaseWrite(&volp->rw);
         if (code == 0) {
@@ -848,8 +877,6 @@ long cm_FindVolumeByName(struct cm_cell *cellp, char *volumeNamep,
     /* if we get here we are holding the mutex */
     if ((volp->flags & CM_VOLUMEFLAG_RESET) && !(flags & CM_GETVOL_FLAG_NO_RESET)) {
         code = cm_UpdateVolumeLocation(cellp, userp, reqp, volp);
-        if (code == 0)
-            volp->flags &= ~CM_VOLUMEFLAG_RESET;
     }	
     lock_ReleaseWrite(&volp->rw);
 
@@ -951,8 +978,6 @@ void cm_ForceUpdateVolume(cm_fid_t *fidp, cm_user_t *userp, cm_req_t *reqp)
      * occur.
      */
     code = cm_UpdateVolumeLocation(cellp, userp, reqp, volp);
-    if (code == 0)
-	volp->flags &= ~CM_VOLUMEFLAG_RESET;
 #endif
     lock_ReleaseWrite(&volp->rw);
 
@@ -1067,7 +1092,7 @@ cm_CheckOfflineVolume(cm_volume_t *volp, afs_uint32 volID)
     char *OfflineMsg;
     char *MOTD;
     cm_req_t req;
-    struct rx_connection * callp;
+    struct rx_connection * rxconnp;
     char volName[32];
     char offLineMsg[256];
     char motd[256];
@@ -1083,8 +1108,6 @@ cm_CheckOfflineVolume(cm_volume_t *volp, afs_uint32 volID)
     if (volp->flags & CM_VOLUMEFLAG_RESET) {
         cm_InitReq(&req);
         code = cm_UpdateVolumeLocation(volp->cellp, cm_rootUserp, &req, volp);
-        if (code == 0)
-            volp->flags &= ~CM_VOLUMEFLAG_RESET;
     }
 
     if (volp->vol[RWVOL].ID != 0 && (!volID || volID == volp->vol[RWVOL].ID) &&
@@ -1106,10 +1129,10 @@ cm_CheckOfflineVolume(cm_volume_t *volp, afs_uint32 volID)
                 if (code) 
                     continue;
 
-                callp = cm_GetRxConn(connp);
-                code = RXAFS_GetVolumeStatus(callp, volp->vol[RWVOL].ID,
+                rxconnp = cm_GetRxConn(connp);
+                code = RXAFS_GetVolumeStatus(rxconnp, volp->vol[RWVOL].ID,
                                              &volStat, &Name, &OfflineMsg, &MOTD);
-                rx_PutConnection(callp);            
+                rx_PutConnection(rxconnp);            
 
             } while (cm_Analyze(connp, cm_rootUserp, &req, NULL, NULL, NULL, NULL, code));
             code = cm_MapRPCError(code, &req);
@@ -1146,10 +1169,10 @@ cm_CheckOfflineVolume(cm_volume_t *volp, afs_uint32 volID)
                 if (code) 
                     continue;
 
-                callp = cm_GetRxConn(connp);
-                code = RXAFS_GetVolumeStatus(callp, volp->vol[ROVOL].ID,
+                rxconnp = cm_GetRxConn(connp);
+                code = RXAFS_GetVolumeStatus(rxconnp, volp->vol[ROVOL].ID,
                                               &volStat, &Name, &OfflineMsg, &MOTD);
-                rx_PutConnection(callp);        
+                rx_PutConnection(rxconnp);        
 
             } while (cm_Analyze(connp, cm_rootUserp, &req, NULL, NULL, NULL, NULL, code));
             code = cm_MapRPCError(code, &req);
@@ -1186,10 +1209,10 @@ cm_CheckOfflineVolume(cm_volume_t *volp, afs_uint32 volID)
                 if (code) 
                     continue;
 
-                callp = cm_GetRxConn(connp);
-                code = RXAFS_GetVolumeStatus(callp, volp->vol[BACKVOL].ID,
+                rxconnp = cm_GetRxConn(connp);
+                code = RXAFS_GetVolumeStatus(rxconnp, volp->vol[BACKVOL].ID,
                                               &volStat, &Name, &OfflineMsg, &MOTD);
-                rx_PutConnection(callp);        
+                rx_PutConnection(rxconnp);        
 
             } while (cm_Analyze(connp, cm_rootUserp, &req, NULL, NULL, NULL, NULL, code));
             code = cm_MapRPCError(code, &req);
@@ -1660,9 +1683,9 @@ cm_VolumeStateByName(cm_volume_t *volp, char *volname)
     size_t len = strlen(volname);
     cm_vol_state_t *statep;
 
-    if (stricmp(".readonly", &volname[len-9]) == 0)
+    if (cm_stricmp_utf8N(".readonly", &volname[len-9]) == 0)
         statep = &volp->vol[ROVOL];
-    else if (stricmp(".backup", &volname[len-7]) == 0)
+    else if (cm_stricmp_utf8N(".backup", &volname[len-7]) == 0)
         statep = &volp->vol[BACKVOL];
     else 
         statep = &volp->vol[RWVOL];
