@@ -10,18 +10,13 @@
 #ifndef __SMB_H_ENV__
 #define __SMB_H_ENV__ 1
 
-#ifdef DJGPP
-#include "netbios95.h"
-#endif /* DJGPP */
-
 #if _WIN32_WINNT < 0x0501
 #undef _WIN32_WINNT
 #define _WIN32_WINNT 0x0501
 #endif
 #include <ntsecapi.h>
 
-/* Support largefiles by default */
-#define AFS_LARGEFILES
+#include <cm_nls.h>
 
 /* basic core protocol SMB structure */
 #pragma pack(push, 1)
@@ -51,7 +46,7 @@ typedef struct smb {
 #define SMB_FLAGS_CANONICAL_PATHNAMES      0x10
 #define SMB_FLAGS_REQUEST_OPLOCK           0x20
 #define SMB_FLAGS_REQUEST_BATCH_OPLOCK     0x40
-#define SMB_FLAGS_SERVER_TO_CLIENT         0x80           
+#define SMB_FLAGS_SERVER_TO_CLIENT         0x80
 
 /* flg2 values */
 
@@ -68,6 +63,7 @@ typedef struct smb {
 
 #define KNOWS_LONG_NAMES(inp) ((((smb_t *)inp)->flg2 & SMB_FLAGS2_KNOWS_LONG_NAMES)?1:0)
 #define WANTS_DFS_PATHNAMES(inp) ((((smb_t *)inp)->flg2 & SMB_FLAGS2_DFS_PATHNAMES)?1:0)
+#define WANTS_UNICODE(inp) ((((smb_t *)inp)->flg2 & SMB_FLAGS2_UNICODE)?1:0)
 
 /* Information Levels */
 #define SMB_INFO_STANDARD               1
@@ -129,6 +125,8 @@ typedef struct smb {
 #define SMB_PACKETSIZE	32768 /* was 8400 */
 /* raw mode is considered obsolete and cannot be used with message signing */
 #define SMB_MAXRAWSIZE  65536
+/* max STRING characters per packet per request */
+#define SMB_STRINGBUFSIZE 4096
 
 /* Negotiate protocol constants */
 /* Security */
@@ -190,10 +188,7 @@ typedef struct smb_packet {
     unsigned char oddByte;
     unsigned short ncb_length;
     unsigned char flags;
-#ifdef DJGPP
-    dos_ptr dos_pkt;
-    unsigned int dos_pkt_sel;
-#endif /* DJGPP */
+    cm_space_t *stringsp;               /* decoded strings from this packet */
 } smb_packet_t;
 
 /* smb_packet flags */
@@ -206,11 +201,6 @@ typedef struct myncb {
     NCB ncb;			        /* ncb to use */
     struct myncb *nextp;		/* when on free list */
     long magic;
-#ifdef DJGPP
-    dos_ptr dos_ncb;
-    smb_packet_t *orig_pkt;
-    unsigned int dos_ncb_sel;
-#endif /* DJGPP */
 } smb_ncb_t;
 
 /* structures representing environments from kernel / SMB network.
@@ -235,7 +225,7 @@ typedef struct smb_vc {
     struct smb_user *usersp;	        /* the first child in the user session list */
     struct smb_fid *fidsp;		/* the first child in the open file list */
     unsigned char errorCount;
-    char rname[17];
+    clientchar_t rname[17];
     int lana;
     char encKey[MSV1_0_CHALLENGE_LENGTH]; /* MSV1_0_CHALLENGE_LENGTH is 8 */
     void * secCtx;                      /* security context when negotiating SMB extended auth
@@ -255,6 +245,7 @@ typedef struct smb_vc {
 #define SMB_VCFLAG_SESSX_RCVD	0x40	/* we received at least one session setups on this vc */
 #define SMB_VCFLAG_AUTH_IN_PROGRESS 0x80 /* a SMB NT extended authentication is in progress */
 #define SMB_VCFLAG_CLEAN_IN_PROGRESS 0x100
+#define SMB_VCFLAG_USEUNICODE   0x200   /* une UNICODE for STRING fields (NTLM 0.12 or later) */
 
 /* one per user session */
 typedef struct smb_user {
@@ -265,7 +256,7 @@ typedef struct smb_user {
     unsigned short userID;		/* the session identifier */
     struct smb_vc *vcp;		        /* back ptr to virtual circuit */
     struct smb_username *unp;           /* user name struct */
-    afs_uint32	delete;			/* ok to del: locked by smb_rctLock */
+    afs_uint32	deleteOk;		/* ok to del: locked by smb_rctLock */
 } smb_user_t;
 
 #define SMB_USERFLAG_DELETE	    1	/* delete struct when ref count zero */
@@ -276,8 +267,8 @@ typedef struct smb_username {
     long flags;			        /* flags; locked by mx */
     osi_mutex_t mx;
     struct cm_user *userp;		/* CM user structure */
-    char *name;			        /* user name */
-    char *machine;                      /* machine name */
+    clientchar_t *name;                 /* user name */
+    clientchar_t *machine;              /* machine name */
     time_t last_logoff_t;		/* most recent logoff time */
 } smb_username_t;
 
@@ -310,8 +301,8 @@ typedef struct smb_tid {
     struct smb_vc *vcp;		        /* back ptr */
     struct cm_user *userp;		/* user logged in at the
 					 * tree connect level (base) */
-    char *pathname;			/* pathname derived from sharename */
-    afs_uint32	delete;			/* ok to del: locked by smb_rctLock */
+    clientchar_t *pathname;             /* pathname derived from sharename */
+    afs_uint32	deleteOk;		/* ok to del: locked by smb_rctLock */
 } smb_tid_t;
 
 #define SMB_TIDFLAG_IPC		1 	/* IPC$ */
@@ -326,36 +317,9 @@ typedef struct smb_pid {
     struct smb_tid *tidp;		/* back ptr */
 } smb_pid_t;
 
-/* ioctl parameter, while being assembled and/or processed */
-typedef struct smb_ioctl {
-    /* input side */
-    char *inDatap;			/* ioctl func's current position
-					 * in input parameter block */
-    char *inAllocp;			/* allocated input parameter block */
-    afs_uint32 inCopied;			/* # of input bytes copied in so far
-					 * by write calls */
-    cm_space_t *prefix;		        /* prefix for subst drives */
-    char *tidPathp;			/* Pathname associated with Tree ID */
 
-    /* output side */
-    char *outDatap;			/* output results assembled so far */
-    char *outAllocp;		        /* output results assembled so far */
-    afs_uint32 outCopied;		/* # of output bytes copied back so far
-                                         * by read calls */
-	
-    /* flags */
-    afs_uint32 flags;
-
-    /* fid pointer */
-    struct smb_fid *fidp;
-
-    /* uid pointer */
-    smb_user_t *uidp;
-} smb_ioctl_t;
-
-/* flags for smb_ioctl_t */
-#define SMB_IOCTLFLAG_DATAIN	1	/* reading data from client to server */
-#define SMB_IOCTLFLAG_LOGON	2	/* got tokens from integrated logon */
+/* Defined in smb_ioctl.h */
+struct smb_ioctl;
 
 /* one per file ID; these are really file descriptors */
 typedef struct smb_fid {
@@ -371,20 +335,20 @@ typedef struct smb_fid {
                                            the file if session is
                                            terminated) */
     osi_hyper_t offset;			/* our file pointer */
-    smb_ioctl_t *ioctlp;		/* ptr to ioctl structure */
+    struct smb_ioctl *ioctlp;		/* ptr to ioctl structure */
 					/* Under NT, we may need to know the
 					 * parent directory and pathname used
 					 * to open the file, either to delete
 					 * the file on close, or to do a
 					 * change notification */
     struct cm_scache *NTopen_dscp;	/* parent directory (NT) */
-    char *NTopen_pathp;		        /* path used in open (NT) */
-    char *NTopen_wholepathp;	        /* entire path, not just last name */
+    clientchar_t *NTopen_pathp;		/* path used in open (NT) */
+    clientchar_t *NTopen_wholepathp;	/* entire path, not just last name */
     int curr_chunk;			/* chunk being read */
     int prev_chunk;			/* previous chunk read */
     int raw_writers;		        /* pending async raw writes */
     EVENT_HANDLE raw_write_event;	/* signal this when raw_writers zero */
-    afs_uint32	delete;			/* ok to del: locked by smb_rctLock */
+    afs_uint32	deleteOk;		/* ok to del: locked by smb_rctLock */
 } smb_fid_t;
 
 #define SMB_FID_OPENREAD_LISTDIR	1	/* open for reading / listing directory */
@@ -460,9 +424,9 @@ typedef struct smb_dirSearch {
 					 * locked by smb_globalLock */
     unsigned short attribute;	        /* search attribute
 					 * (used for extended protocol) */
-    char tidPath[256];                  /* tid path */
-    char relPath[1024];                 /* relative path */        
-    char mask[256];			/* search mask for V3 */
+    clientchar_t tidPath[256];          /* tid path */
+    clientchar_t relPath[1024];         /* relative path */        
+    clientchar_t mask[256];	        /* search mask for V3 */
 } smb_dirSearch_t;
 
 #define SMB_DIRSEARCH_DELETE	1	/* delete struct when ref count zero */
@@ -473,7 +437,7 @@ typedef struct smb_dirSearch {
 /* type for patching directory listings */
 typedef struct smb_dirListPatch {
     osi_queue_t q;
-    char *dptr;		                /* ptr to attr, time, data, sizel, sizeh */
+    char *dptr;                      /* ptr to attr, time, data, sizel, sizeh */
     long flags;                         /* flags.  See below */
     cm_fid_t fid;
     cm_dirEntry_t *dep;                 /* temp */
@@ -532,12 +496,7 @@ typedef struct smb_dispatch {
 
 /* prototypes */
 
-extern void smb_Init(osi_log_t *logp, int useV3,
-	int nThreads
-#ifndef DJGPP
-        , void *aMBfunc
-#endif
-  );
+extern void smb_Init(osi_log_t *logp, int useV3, int nThreads, void *aMBfunc);
 
 extern void smb_LargeSearchTimeFromUnixTime(FILETIME *largeTimep, time_t unixTime);
 
@@ -569,9 +528,11 @@ extern void smb_ReleaseTID(smb_tid_t *tidp, afs_uint32 locked);
 
 extern smb_user_t *smb_FindUID(smb_vc_t *vcp, unsigned short uid, int flags);
 
-extern smb_username_t *smb_FindUserByName(char *usern, char *machine, afs_uint32 flags);
+extern smb_username_t *smb_FindUserByName(clientchar_t *usern, clientchar_t *machine, afs_uint32 flags);
 
-extern smb_user_t *smb_FindUserByNameThisSession(smb_vc_t *vcp, char *usern); 
+extern cm_user_t *smb_FindCMUserByName(clientchar_t *usern, clientchar_t *machine, afs_uint32 flags);
+
+extern smb_user_t *smb_FindUserByNameThisSession(smb_vc_t *vcp, clientchar_t *usern); 
 
 extern void smb_ReleaseUsername(smb_username_t *unp);
 
@@ -583,7 +544,7 @@ extern cm_user_t *smb_GetUserFromVCP(smb_vc_t *vcp, smb_packet_t *inp);
 
 extern cm_user_t *smb_GetUserFromUID(smb_user_t *uidp);
 
-extern long smb_LookupTIDPath(smb_vc_t *vcp, unsigned short tid, char ** tidPathp);
+extern long smb_LookupTIDPath(smb_vc_t *vcp, unsigned short tid, clientchar_t ** tidPathp);
 
 extern smb_fid_t *smb_FindFID(smb_vc_t *vcp, unsigned short fid, int flags);
 
@@ -596,9 +557,9 @@ extern void smb_ReleaseFID(smb_fid_t *fidp);
 extern long smb_CloseFID(smb_vc_t *vcp, smb_fid_t *fidp, cm_user_t *userp,
                          afs_uint32 dosTime);
 
-extern int smb_FindShare(smb_vc_t *vcp, smb_user_t *uidp, char *shareName, char **pathNamep);
+extern int smb_FindShare(smb_vc_t *vcp, smb_user_t *uidp, clientchar_t *shareName, clientchar_t **pathNamep);
 
-extern int smb_FindShareCSCPolicy(char *shareName);
+extern int smb_FindShareCSCPolicy(clientchar_t *shareName);
 
 extern smb_dirSearch_t *smb_FindDirSearchNL(long cookie);
 
@@ -634,13 +595,36 @@ extern void smb_SetSMBParmDouble(smb_packet_t *smbp, int slot, char *parmValuep)
 
 extern void smb_SetSMBParmByte(smb_packet_t *smbp, int slot, unsigned int parmValue);
 
-extern void smb_StripLastComponent(char *outPathp, char **lastComponentp,
-	char *inPathp);
+extern void smb_StripLastComponent(clientchar_t *outPathp, clientchar_t **lastComponentp,
+	clientchar_t *inPathp);
 
-extern unsigned char *smb_ParseASCIIBlock(unsigned char *inp, char **chainpp);
+#define SMB_STRF_FORCEASCII (1<<0)
+#define SMB_STRF_ANSIPATH   (1<<1)
+#define SMB_STRF_IGNORENULL (1<<2)
+
+extern clientchar_t *smb_ParseASCIIBlock(smb_packet_t * pktp, unsigned char *inp,
+                                         char **chainpp, int flags);
+
+extern clientchar_t *smb_ParseString(smb_packet_t * pktp, unsigned char * inp,
+                                     char ** chainpp, int flags);
+
+extern clientchar_t *smb_ParseStringBuf(const unsigned char * bufbase,
+                                        cm_space_t ** stringspp,
+                                        unsigned char *inp, size_t *pcb_max,
+                                        char **chainpp, int flags);
+
+extern clientchar_t *smb_ParseStringCb(smb_packet_t * pktp, unsigned char * inp,
+                                       size_t cb, char ** chainpp, int flags);
+
+extern clientchar_t *smb_ParseStringCch(smb_packet_t * pktp, unsigned char * inp,
+                                        size_t cch, char ** chainpp, int flags);
+
+extern unsigned char * smb_UnparseString(smb_packet_t * pktp, unsigned char * outp,
+                                   clientchar_t * str,
+                                   size_t * plen, int flags);
 
 extern unsigned char *smb_ParseVblBlock(unsigned char *inp, char **chainpp,
-	int *lengthp);
+                                  int *lengthp);
 
 extern smb_packet_t *smb_GetResponsePacket(smb_vc_t *vcp, smb_packet_t *inp);
 
@@ -672,7 +656,7 @@ extern int smb_maxMpxRequests; /* max # of mpx requests */
 
 extern int smb_StoreAnsiFilenames;
 extern int smb_hideDotFiles;
-extern unsigned int smb_IsDotFile(char *lastComp);
+extern unsigned int smb_IsDotFile(clientchar_t *lastComp);
 extern afs_uint32 smb_AsyncStore;
 extern afs_uint32 smb_AsyncStoreSize;
 
@@ -685,14 +669,15 @@ extern int smb_authType; /* Type of SMB authentication to be used. One from belo
 
 extern HANDLE smb_lsaHandle; /* LSA handle obtained during smb_init if using SMB auth */
 extern ULONG smb_lsaSecPackage; /* LSA security package id. Set during smb_init */
-extern char smb_ServerDomainName[];
+extern clientchar_t smb_ServerDomainName[];
 extern int smb_ServerDomainNameLength;
-extern char smb_ServerOS[];
+extern clientchar_t smb_ServerOS[];
 extern int smb_ServerOSLength;
-extern char smb_ServerLanManager[];
+extern clientchar_t smb_ServerLanManager[];
 extern int smb_ServerLanManagerLength;
 extern GUID smb_ServerGUID;
 extern LSA_STRING smb_lsaLogonOrigin;
+extern LONG smb_UseUnicode;
 
 /* used for getting a challenge for SMB auth */
 typedef struct _MSV1_0_LM20_CHALLENGE_REQUEST {  
@@ -705,9 +690,9 @@ typedef struct _MSV1_0_LM20_CHALLENGE_RESPONSE {
 } MSV1_0_LM20_CHALLENGE_RESPONSE, *PMSV1_0_LM20_CHALLENGE_RESPONSE;
 /**/
 
-extern long smb_AuthenticateUserLM(smb_vc_t *vcp, char * accountName, char * primaryDomain, char * ciPwd, unsigned ciPwdLength, char * csPwd, unsigned csPwdLength);
+extern long smb_AuthenticateUserLM(smb_vc_t *vcp, clientchar_t * accountName, clientchar_t * primaryDomain, char * ciPwd, unsigned ciPwdLength, char * csPwd, unsigned csPwdLength);
 
-extern long smb_GetNormalizedUsername(char * usern, const char * accountName, const char * domainName);
+extern long smb_GetNormalizedUsername(clientchar_t * usern, const clientchar_t * accountName, const clientchar_t * domainName);
 
 extern void smb_FormatResponsePacket(smb_vc_t *vcp, smb_packet_t *inp,
 	smb_packet_t *op);
@@ -724,33 +709,21 @@ extern int smb_ChainFID(int fid, smb_packet_t *inp);
 
 extern unsigned char *smb_ParseDataBlock(unsigned char *inp, char **chainpp, int *lengthp);
 
-extern unsigned char *smb_ParseASCIIBlock(unsigned char *inp, char **chainpp);
-
 extern unsigned char *smb_ParseVblBlock(unsigned char *inp, char **chainpp, int *lengthp);
 
 extern int smb_SUser(cm_user_t *userp);
 
-#ifndef DJGPP
 long smb_WriteData(smb_fid_t *fidp, osi_hyper_t *offsetp, afs_uint32 count, char *op,
 	cm_user_t *userp, long *writtenp);
-#else /* DJGPP */
-long smb_WriteData(smb_fid_t *fidp, osi_hyper_t *offsetp, afs_uint32 count, char *op,
-	cm_user_t *userp, long *writtenp, int dosflag);
-#endif /* !DJGPP */
 
-#ifndef DJGPP
 extern long smb_ReadData(smb_fid_t *fidp, osi_hyper_t *offsetp, afs_uint32 count,
 	char *op, cm_user_t *userp, long *readp);
-#else /* DJGPP */
-extern long smb_ReadData(smb_fid_t *fidp, osi_hyper_t *offsetp, afs_uint32 count,
-	char *op, cm_user_t *userp, long *readp, int dosflag);
-#endif /* !DJGPP */
 
-extern long smb_Rename(smb_vc_t *vcp, smb_packet_t *inp, char *oldPathp, char *newPathp, int attrs);
+extern long smb_Rename(smb_vc_t *vcp, smb_packet_t *inp, clientchar_t *oldPathp, clientchar_t *newPathp, int attrs);
 
-extern long smb_Link(smb_vc_t *vcp, smb_packet_t *inp, char *oldPathp, char *newPathp);
+extern long smb_Link(smb_vc_t *vcp, smb_packet_t *inp, clientchar_t *oldPathp, clientchar_t *newPathp);
 
-extern BOOL smb_IsLegalFilename(char *filename);
+extern BOOL smb_IsLegalFilename(clientchar_t *filename);
 
 extern char *smb_GetSharename(void);
 
@@ -766,6 +739,7 @@ extern long smb_IsNetworkStarted(void);
 extern void smb_LanAdapterChange(int);
 extern void smb_SetLanAdapterChangeDetected(void);
 
+extern void smb_InitReq(cm_req_t *reqp);
 
 #define SMB_LISTENER_UNINITIALIZED -1
 #define SMB_LISTENER_STOPPED 0
@@ -776,7 +750,11 @@ extern void smb_SetLanAdapterChangeDetected(void);
 #include "smb_ioctl.h"
 #include "smb_iocons.h"
 
-cm_user_t *smb_FindOrCreateUser(smb_vc_t *vcp, char *usern);
+cm_user_t *smb_FindOrCreateUser(smb_vc_t *vcp, clientchar_t *usern);
+
+int smb_DumpVCP(FILE *outputFile, char *cookie, int lock);
+
+void smb_Shutdown(void);
 
 #ifdef NOTSERVICE
 extern void smb_LogPacket(smb_packet_t *packet);
