@@ -23,7 +23,7 @@
 #include "afs/param.h"
 
 RCSID
-    ("$Header: /cvs/openafs/src/afs/VNOPS/afs_vnop_access.c,v 1.11.8.3 2006/07/31 21:27:40 shadow Exp $");
+    ("$Header: /cvs/openafs/src/afs/VNOPS/afs_vnop_access.c,v 1.11.8.6 2008/05/23 14:25:16 shadow Exp $");
 
 #include "afs/sysincludes.h"	/* Standard vendor system headers */
 #include "afsincludes.h"	/* Afs-based standard headers */
@@ -209,24 +209,50 @@ afs_access(OSI_VC_DECL(avc), register afs_int32 amode,
     if ((code = afs_InitReq(&treq, acred)))
 	return code;
 
-    code = afs_EvalFakeStat(&avc, &fakestate, &treq);
+    AFS_DISCON_LOCK();
+
+    if (afs_fakestat_enable && avc->mvstat == 1) {
+	code = afs_TryEvalFakeStat(&avc, &fakestate, &treq);
+        if (code == 0 && avc->mvstat == 1) {
+	    afs_PutFakeStat(&fakestate);
+	    AFS_DISCON_UNLOCK();
+	    return 0;
+        }
+    } else {
+	code = afs_EvalFakeStat(&avc, &fakestate, &treq);
+    }
+
     if (code) {
 	afs_PutFakeStat(&fakestate);
+	AFS_DISCON_UNLOCK();
 	return code;
     }
 
-    code = afs_VerifyVCache(avc, &treq);
-    if (code) {
-	afs_PutFakeStat(&fakestate);
-	code = afs_CheckCode(code, &treq, 16);
-	return code;
+    if (vType(avc) != VDIR || !afs_InReadDir(avc)) {
+	code = afs_VerifyVCache(avc, &treq);
+	if (code) {
+	    afs_PutFakeStat(&fakestate);
+	    AFS_DISCON_UNLOCK();
+	    code = afs_CheckCode(code, &treq, 16);
+	    return code;
+	}
     }
 
     /* if we're looking for write access and we have a read-only file system, report it */
     if ((amode & VWRITE) && (avc->states & CRO)) {
 	afs_PutFakeStat(&fakestate);
+	AFS_DISCON_UNLOCK();
 	return EROFS;
     }
+    
+    /* If we're looking for write access, and we're disconnected without logging, forget it */
+    if ((amode & VWRITE) && (AFS_IS_DISCONNECTED && !AFS_IS_LOGGING)) {
+        afs_PutFakeStat(&fakestate);
+	AFS_DISCON_UNLOCK();
+	/*printf("Network is down in afs_vnop_access\n");*/
+        return ENETDOWN;
+    }
+    
     code = 1;			/* Default from here on in is access ok. */
     if (avc->states & CForeign) {
 	/* In the dfs xlator the EXEC bit is mapped to LOOKUP */
@@ -305,6 +331,9 @@ afs_access(OSI_VC_DECL(avc), register afs_int32 amode,
 	}
     }
     afs_PutFakeStat(&fakestate);
+
+    AFS_DISCON_UNLOCK();
+    
     if (code) {
 	return 0;		/* if access is ok */
     } else {
