@@ -11,7 +11,7 @@
 #include <afs/param.h>
 
 RCSID
-    ("$Header: /cvs/openafs/src/ptserver/pts.c,v 1.15.2.2 2007/08/11 23:41:37 jaltman Exp $");
+    ("$Header: /cvs/openafs/src/ptserver/pts.c,v 1.15.2.6 2008/04/02 19:51:56 shadow Exp $");
 
 #include <stdio.h>
 #include <string.h>
@@ -50,22 +50,29 @@ struct sourcestack {
     FILE *s_file;
 } *shead;
 
+struct authstate {
+    int sec;
+    const char *confdir;
+    char cell[MAXCELLCHARS];
+};
+
 int
-pts_Interactive(struct cmd_syndesc *as, char *arock)
+pts_Interactive(struct cmd_syndesc *as, void *arock)
 {
+    source = stdin;
     finished = 0;
     return 0;
 }
 
 int
-pts_Quit(struct cmd_syndesc *as, char *arock)
+pts_Quit(struct cmd_syndesc *as, void *arock)
 {
     finished = 1;
     return 0;
 }
 
 int
-pts_Source(struct cmd_syndesc *as, char *arock)
+pts_Source(struct cmd_syndesc *as, void *arock)
 {
     FILE *fd;
     struct sourcestack *sp;
@@ -93,7 +100,7 @@ pts_Source(struct cmd_syndesc *as, char *arock)
 }
 
 int
-pts_Sleep(struct cmd_syndesc *as, char *arock)
+pts_Sleep(struct cmd_syndesc *as, void *arock)
 {
     int delay;
     if (!as->parms[0].items) {
@@ -101,7 +108,11 @@ pts_Sleep(struct cmd_syndesc *as, char *arock)
 	return 1;
     }
     delay = atoi(as->parms[0].items->data);
+#ifdef AFS_PTHREAD_ENV
+    sleep(delay);
+#else
     IOMGR_Sleep(delay);
+#endif
     return 0;
 }
 
@@ -130,39 +141,78 @@ osi_audit()
 }
 
 int
-GetGlobals(struct cmd_syndesc *as, char *arock)
+GetGlobals(struct cmd_syndesc *as, void *arock)
 {
-    register afs_int32 code;
-    char *cell;
-    afs_int32 sec = 1;
+    struct authstate *state = (struct authstate *) arock;
+    afs_int32 code;
+    char *cell = NULL;
+    afs_int32 sec;
+    int changed = 0;
+    const char* confdir;
 
     whoami = as->a0name;
 
     if (!strcmp(as->name, "help"))
 	return 0;
-    if (as->parms[16].items)
-	cell = as->parms[16].items->data;
-    else
-	cell = 0;
-    if (as->parms[17].items)
-	sec = 0;
 
-    if (as->parms[18].items) {	/* testing? */
-	code = pr_Initialize(sec, AFSDIR_SERVER_ETC_DIRPATH, cell);
+    if (*state->cell) {
+	cell = state->cell;
+    }
+    sec = state->sec;
+
+    if (state->confdir == NULL) {
+	changed = 1;
+    }
+
+    if (as->parms[16].items) {
+	changed = 1;
+	cell = as->parms[16].items->data;
+    }
+    if (as->parms[17].items) { /* -noauth */
+	changed = 1;
+	sec = 0;
+    }
+    if (as->parms[20].items) { /* -localauth */
+	changed = 1;
+	sec = 2;
+    }
+    if (as->parms[21].items) { /* -auth */
+	changed = 1;
+	sec = 1;
+    }
+    if (as->parms[18].items || as->parms[20].items) { /* -test, -localauth */
+	changed = 1;
+	confdir = AFSDIR_SERVER_ETC_DIRPATH;
     } else {
-	code = pr_Initialize(sec, AFSDIR_CLIENT_ETC_DIRPATH, cell);
+	if (sec == 2)
+	    confdir = AFSDIR_SERVER_ETC_DIRPATH;
+	else
+	    confdir = AFSDIR_CLIENT_ETC_DIRPATH;
+    }
+    if (changed) {
+	CleanUp(as, arock);
+	code = pr_Initialize(sec, confdir, cell);
+    } else {
+	code = 0;
     }
     if (code) {
 	afs_com_err(whoami, code, "while initializing");
 	return code;
     }
+    state->sec = sec;
+    state->confdir = confdir;
+    if (cell && cell != state->cell)
+        strncpy(state->cell, cell, MAXCELLCHARS-1);
+
+    force = 0;
     if (as->parms[19].items)
 	force = 1;
+
     return code;
 }
 
 int
-CleanUp(struct cmd_syndesc *as, char *arock)
+CleanUp(struct cmd_syndesc *as, void *arock)
 {
     if (as && !strcmp(as->name, "help"))
 	return 0;
@@ -175,7 +225,7 @@ CleanUp(struct cmd_syndesc *as, char *arock)
 }
 
 int
-CreateGroup(struct cmd_syndesc *as, char *arock)
+CreateGroup(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     afs_int32 id;
@@ -235,7 +285,7 @@ CreateGroup(struct cmd_syndesc *as, char *arock)
 }
 
 int
-CreateUser(struct cmd_syndesc *as, char *arock)
+CreateUser(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     afs_int32 id;
@@ -460,7 +510,7 @@ GetNameOrId(register struct cmd_syndesc *as, struct idlist *lids, struct namelis
 
 
 int
-AddToGroup(struct cmd_syndesc *as, char *arock)
+AddToGroup(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     struct cmd_item *u, *g;
@@ -481,7 +531,7 @@ AddToGroup(struct cmd_syndesc *as, char *arock)
 }
 
 int
-RemoveFromGroup(struct cmd_syndesc *as, char *arock)
+RemoveFromGroup(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     struct cmd_item *u, *g;
@@ -502,7 +552,7 @@ RemoveFromGroup(struct cmd_syndesc *as, char *arock)
 }
 
 int
-ListMembership(struct cmd_syndesc *as, char *arock)
+ListMembership(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     idlist ids;
@@ -547,7 +597,7 @@ ListMembership(struct cmd_syndesc *as, char *arock)
 }
 
 int
-Delete(struct cmd_syndesc *as, char *arock)
+Delete(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     idlist ids;
@@ -586,7 +636,7 @@ char *flags_dncase = "s mar";	/* legal member acces values */
 int flags_shift[5] = { 2, 1, 2, 2, 1 };	/* bits for each */
 
 int
-CheckEntry(struct cmd_syndesc *as, char *arock)
+CheckEntry(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     afs_int32 rcode = 1;
@@ -685,7 +735,7 @@ CheckEntry(struct cmd_syndesc *as, char *arock)
 }
 
 int
-ListEntries(struct cmd_syndesc *as, char *arock)
+ListEntries(struct cmd_syndesc *as, void *arock)
 {
     afs_int32 code = 0;
     afs_int32 flag, startindex, nentries, nextstartindex;
@@ -704,7 +754,7 @@ ListEntries(struct cmd_syndesc *as, char *arock)
 	    pr_ListEntries(flag, startindex, &nentries, &entriesp,
 			   &nextstartindex);
 	if (code) {
-	    afs_com_err(whoami, code, "; unable to list entries\n");
+	    afs_com_err(whoami, code, "; unable to list entries");
 	    if (entriesp)
 		free(entriesp);
 	    break;
@@ -722,7 +772,7 @@ ListEntries(struct cmd_syndesc *as, char *arock)
 }
 
 int
-ChownGroup(struct cmd_syndesc *as, char *arock)
+ChownGroup(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     char *name;
@@ -738,7 +788,7 @@ ChownGroup(struct cmd_syndesc *as, char *arock)
 }
 
 int
-ChangeName(struct cmd_syndesc *as, char *arock)
+ChangeName(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     char *oldname;
@@ -754,7 +804,7 @@ ChangeName(struct cmd_syndesc *as, char *arock)
 }
 
 int
-ListMax(struct cmd_syndesc *as, char *arock)
+ListMax(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     afs_int32 maxUser, maxGroup;
@@ -775,7 +825,7 @@ ListMax(struct cmd_syndesc *as, char *arock)
 }
 
 int
-SetMax(struct cmd_syndesc *as, char *arock)
+SetMax(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     afs_int32 maxid;
@@ -815,7 +865,7 @@ SetMax(struct cmd_syndesc *as, char *arock)
 }
 
 int
-SetFields(struct cmd_syndesc *as, char *arock)
+SetFields(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     idlist ids;
@@ -910,7 +960,7 @@ SetFields(struct cmd_syndesc *as, char *arock)
 }
 
 int
-ListOwned(struct cmd_syndesc *as, char *arock)
+ListOwned(struct cmd_syndesc *as, void *arock)
 {
     register afs_int32 code;
     idlist ids;
@@ -972,6 +1022,10 @@ add_std_args(register struct cmd_syndesc *ts)
     cmd_AddParm(ts, "-test", CMD_FLAG, CMD_OPTIONAL | CMD_HIDE, test_help);
     cmd_AddParm(ts, "-force", CMD_FLAG, CMD_OPTIONAL,
 		"Continue oper despite reasonable errors");
+    cmd_AddParm(ts, "-localauth", CMD_FLAG, CMD_OPTIONAL,
+		"use local authentication");
+    cmd_AddParm(ts, "-auth", CMD_FLAG, CMD_OPTIONAL,
+		"use user's authentication (default)");
 }
 
 /*
@@ -996,6 +1050,7 @@ main(int argc, char **argv)
     int parsec;
     char *parsev[CMD_MAXPARMS];
     char *savec;
+    struct authstate state;
 
 #ifdef WIN32
     WSADATA WSAjunk;
@@ -1020,7 +1075,10 @@ main(int argc, char **argv)
     sigaction(SIGSEGV, &nsa, NULL);
 #endif
 
-    ts = cmd_CreateSyntax("creategroup", CreateGroup, 0,
+    memset(&state, 0, sizeof(state));
+    state.sec = 1; /* default is auth */
+
+    ts = cmd_CreateSyntax("creategroup", CreateGroup, NULL,
 			  "create a new group");
     cmd_AddParm(ts, "-name", CMD_LIST, 0, "group name");
     cmd_AddParm(ts, "-owner", CMD_SINGLE, CMD_OPTIONAL, "owner of the group");
@@ -1029,60 +1087,60 @@ main(int argc, char **argv)
     add_std_args(ts);
     cmd_CreateAlias(ts, "cg");
 
-    ts = cmd_CreateSyntax("createuser", CreateUser, 0, "create a new user");
+    ts = cmd_CreateSyntax("createuser", CreateUser, NULL, "create a new user");
     cmd_AddParm(ts, "-name", CMD_LIST, 0, "user name");
     cmd_AddParm(ts, "-id", CMD_LIST, CMD_OPTIONAL, "user id");
     add_std_args(ts);
     cmd_CreateAlias(ts, "cu");
 
-    ts = cmd_CreateSyntax("adduser", AddToGroup, 0, "add a user to a group");
+    ts = cmd_CreateSyntax("adduser", AddToGroup, NULL, "add a user to a group");
     cmd_AddParm(ts, "-user", CMD_LIST, 0, "user name");
     cmd_AddParm(ts, "-group", CMD_LIST, 0, "group name");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("removeuser", RemoveFromGroup, 0,
+    ts = cmd_CreateSyntax("removeuser", RemoveFromGroup, NULL,
 			  "remove a user from a group");
     cmd_AddParm(ts, "-user", CMD_LIST, 0, "user name");
     cmd_AddParm(ts, "-group", CMD_LIST, 0, "group name");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("membership", ListMembership, 0,
+    ts = cmd_CreateSyntax("membership", ListMembership, NULL,
 			  "list membership of a user or group");
     cmd_AddParm(ts, "-nameorid", CMD_LIST, 0, "user or group name or id");
     add_std_args(ts);
     cmd_CreateAlias(ts, "groups");
 
-    ts = cmd_CreateSyntax("delete", Delete, 0,
+    ts = cmd_CreateSyntax("delete", Delete, NULL,
 			  "delete a user or group from database");
     cmd_AddParm(ts, "-nameorid", CMD_LIST, 0, "user or group name or id");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("examine", CheckEntry, 0, "examine an entry");
+    ts = cmd_CreateSyntax("examine", CheckEntry, NULL, "examine an entry");
     cmd_AddParm(ts, "-nameorid", CMD_LIST, 0, "user or group name or id");
     add_std_args(ts);
     cmd_CreateAlias(ts, "check");
 
-    ts = cmd_CreateSyntax("chown", ChownGroup, 0,
+    ts = cmd_CreateSyntax("chown", ChownGroup, NULL,
 			  "change ownership of a group");
     cmd_AddParm(ts, "-name", CMD_SINGLE, 0, "group name");
     cmd_AddParm(ts, "-owner", CMD_SINGLE, 0, "new owner");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("rename", ChangeName, 0, "rename user or group");
+    ts = cmd_CreateSyntax("rename", ChangeName, NULL, "rename user or group");
     cmd_AddParm(ts, "-oldname", CMD_SINGLE, 0, "old name");
     cmd_AddParm(ts, "-newname", CMD_SINGLE, 0, "new name");
     add_std_args(ts);
     cmd_CreateAlias(ts, "chname");
 
-    ts = cmd_CreateSyntax("listmax", ListMax, 0, "list max id");
+    ts = cmd_CreateSyntax("listmax", ListMax, NULL, "list max id");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("setmax", SetMax, 0, "set max id");
+    ts = cmd_CreateSyntax("setmax", SetMax, NULL, "set max id");
     cmd_AddParm(ts, "-group", CMD_SINGLE, CMD_OPTIONAL, "group max");
     cmd_AddParm(ts, "-user", CMD_SINGLE, CMD_OPTIONAL, "user max");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("setfields", SetFields, 0,
+    ts = cmd_CreateSyntax("setfields", SetFields, NULL,
 			  "set fields for an entry");
     cmd_AddParm(ts, "-nameorid", CMD_LIST, 0, "user or group name or id");
     cmd_AddParm(ts, "-access", CMD_SINGLE, CMD_OPTIONAL, "set privacy flags");
@@ -1094,42 +1152,42 @@ main(int argc, char **argv)
 #endif
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("listowned", ListOwned, 0,
+    ts = cmd_CreateSyntax("listowned", ListOwned, NULL,
 			  "list groups owned by an entry or zero id gets orphaned groups");
     cmd_AddParm(ts, "-nameorid", CMD_LIST, 0, "user or group name or id");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("listentries", ListEntries, 0,
+    ts = cmd_CreateSyntax("listentries", ListEntries, NULL,
 			  "list users/groups in the protection database");
     cmd_AddParm(ts, "-users", CMD_FLAG, CMD_OPTIONAL, "list user entries");
     cmd_AddParm(ts, "-groups", CMD_FLAG, CMD_OPTIONAL, "list group entries");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("interactive", pts_Interactive, 0,
+    ts = cmd_CreateSyntax("interactive", pts_Interactive, NULL,
 			  "enter interactive mode");
     add_std_args(ts);
     cmd_CreateAlias(ts, "in");
 
-    ts = cmd_CreateSyntax("quit", pts_Quit, 0, "exit program");
+    ts = cmd_CreateSyntax("quit", pts_Quit, NULL, "exit program");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("source", pts_Source, 0, "read commands from file");
+    ts = cmd_CreateSyntax("source", pts_Source, NULL, "read commands from file");
     cmd_AddParm(ts, "-file", CMD_SINGLE, 0, "filename");
     add_std_args(ts);
 
-    ts = cmd_CreateSyntax("sleep", pts_Sleep, 0, "pause for a bit");
+    ts = cmd_CreateSyntax("sleep", pts_Sleep, NULL, "pause for a bit");
     cmd_AddParm(ts, "-delay", CMD_SINGLE, 0, "seconds");
     add_std_args(ts);
 
-    cmd_SetBeforeProc(GetGlobals, 0);
+    cmd_SetBeforeProc(GetGlobals, &state);
 
     finished = 1;
-    source = stdin;
+    source = NULL;
     if (code = cmd_Dispatch(argc, argv)) {
 	CleanUp(NULL, NULL);
 	exit(1);
     }
-    while (!finished) {
+    while (source && !finished) {
 	if (isatty(fileno(source)))
 	    fprintf(stderr, "pts> ");
 	if (!fgets(line, sizeof line, source)) {
