@@ -6,7 +6,7 @@
  * License.  For details, see the LICENSE file in the top-level source
  * directory or online at http://www.openafs.org/dl/license10.html
  *
- * Portions Copyright (c) 2006 Sine Nomine Associates
+ * Portions Copyright (c) 2006-2008 Sine Nomine Associates
  */
 
 /*
@@ -29,6 +29,38 @@ typedef bit32 FileOffset;	/* Offset in this file */
 #include "daemon_com.h"
 #include "fssync.h"
 
+#if 0
+/** turn this on if you suspect a volume package locking bug */
+#define VOL_LOCK_DEBUG 1
+#endif
+
+#ifdef VOL_LOCK_DEBUG
+#define VOL_LOCK_ASSERT_HELD \
+    assert(vol_glock_holder == pthread_self())
+#define VOL_LOCK_ASSERT_UNHELD \
+    assert(vol_glock_holder == 0)
+#define _VOL_LOCK_SET_HELD \
+    vol_glock_holder = pthread_self()
+#define _VOL_LOCK_SET_UNHELD \
+    vol_glock_holder = 0
+#define VOL_LOCK_DBG_CV_WAIT_END \
+    do { \
+        VOL_LOCK_ASSERT_UNHELD; \
+        _VOL_LOCK_SET_HELD; \
+    } while(0)
+#define VOL_LOCK_DBG_CV_WAIT_BEGIN \
+    do { \
+         VOL_LOCK_ASSERT_HELD; \
+         _VOL_LOCK_SET_UNHELD; \
+    } while(0)
+#else
+#define VOL_LOCK_ASSERT_HELD
+#define VOL_LOCK_ASSERT_UNHELD
+#define VOL_LOCK_DBG_CV_WAIT_BEGIN
+#define VOL_LOCK_DBG_CV_WAIT_END
+#endif
+
+
 #ifdef AFS_PTHREAD_ENV
 #include <assert.h>
 #include <pthread.h>
@@ -37,10 +69,33 @@ extern pthread_mutex_t vol_trans_mutex;
 extern pthread_cond_t vol_put_volume_cond;
 extern pthread_cond_t vol_sleep_cond;
 extern int vol_attach_threads;
+#ifdef VOL_LOCK_DEBUG
+extern pthread_t vol_glock_holder;
+#define VOL_LOCK \
+    do { \
+        assert(pthread_mutex_lock(&vol_glock_mutex) == 0); \
+        assert(vol_glock_holder == 0); \
+        vol_glock_holder = pthread_self(); \
+    } while (0)
+#define VOL_UNLOCK \
+    do { \
+        VOL_LOCK_ASSERT_HELD; \
+        vol_glock_holder = 0; \
+        assert(pthread_mutex_unlock(&vol_glock_mutex) == 0); \
+    } while (0)
+#define VOL_CV_WAIT(cv) \
+    do { \
+        VOL_LOCK_DBG_CV_WAIT_BEGIN; \
+        assert(pthread_cond_wait((cv), &vol_glock_mutex) == 0); \
+        VOL_LOCK_DBG_CV_WAIT_END; \
+    } while (0)
+#else /* !VOL_LOCK_DEBUG */
 #define VOL_LOCK \
     assert(pthread_mutex_lock(&vol_glock_mutex) == 0)
 #define VOL_UNLOCK \
     assert(pthread_mutex_unlock(&vol_glock_mutex) == 0)
+#define VOL_CV_WAIT(cv) assert(pthread_cond_wait((cv), &vol_glock_mutex) == 0)
+#endif /* !VOL_LOCK_DEBUG */
 #define VSALVSYNC_LOCK \
     assert(pthread_mutex_lock(&vol_salvsync_mutex) == 0)
 #define VSALVSYNC_UNLOCK \
@@ -58,11 +113,16 @@ extern int vol_attach_threads;
 #define VTRANS_UNLOCK
 #endif /* AFS_PTHREAD_ENV */
 
-typedef enum { fileServer,       /* the fileserver process */
-	       volumeUtility,    /* volserver, or a single volume salvager (non-dafs) */
-	       salvager,         /* standalone whole-partition salvager */
-	       salvageServer,    /* dafs online salvager */
-	       debugUtility      /* fssync-debug or similar utility */
+/**
+ * volume package program type enumeration.
+ */
+typedef enum {
+    fileServer          = 1,    /**< the fileserver process */
+    volumeUtility       = 2,    /**< volserver, or a 
+				 *   single volume salvager (non-dafs) */
+    salvager            = 3,    /**< standalone whole-partition salvager */
+    salvageServer       = 4,    /**< dafs online salvager */
+    debugUtility        = 5     /**< fssync-debug or similar utility */
 } ProgramType;
 extern ProgramType programType;	/* The type of program using the package */
 
@@ -72,6 +132,15 @@ extern int (*V_BreakVolumeCallbacks) ();
 extern int (*vol_PollProc) ();
 #define	DOPOLL	((vol_PollProc)? (*vol_PollProc)() : 0)
 
+#ifdef AFS_DEMAND_ATTACH_FS
+/**
+ * variable error return code based upon programType and DAFS presence
+ */
+#define DAFS_VSALVAGE   ((programType == fileServer) ? VSALVAGING : VSALVAGE)
+#else
+#define DAFS_VSALVAGE   (VSALVAGE)
+#endif
+
 struct versionStamp {		/* Version stamp for critical volume files */
     bit32 magic;		/* Magic number */
     bit32 version;		/* Version number of this file, or software
@@ -79,45 +148,55 @@ struct versionStamp {		/* Version stamp for critical volume files */
 };
 
 #ifdef AFS_DEMAND_ATTACH_FS
-/*
- * demand attach fs
- * volume state machine
+/**
+ * demand attach volume state enumeration.
  *
- * these must be contiguous in order for IsValidState() to work correctly 
+ * @note values must be contiguous in order for VIsValidState() to work correctly
  */
-#define VOL_STATE_UNATTACHED     0       /* volume is unattached */
-#define VOL_STATE_PREATTACHED    1       /* volume has been pre-attached */
-#define VOL_STATE_ATTACHING      2       /* volume is transitioning to fully attached */
-#define VOL_STATE_ATTACHED       3       /* volume has been fully attached */
-#define VOL_STATE_UPDATING       4       /* volume is updating on-disk structures */
-#define VOL_STATE_GET_BITMAP     5       /* volume is getting bitmap entries */
-#define VOL_STATE_HDR_LOADING    6       /* volume is loading disk header */
-#define VOL_STATE_HDR_ATTACHING  7       /* volume is getting a header from the LRU */
-#define VOL_STATE_SHUTTING_DOWN  8       /* volume is shutting down */
-#define VOL_STATE_GOING_OFFLINE  9       /* volume is going offline */
-#define VOL_STATE_OFFLINING      10      /* volume is transitioning to offline */
-#define VOL_STATE_DETACHING      11      /* volume is transitioning to detached */
-#define VOL_STATE_SALVSYNC_REQ   12      /* volume is blocked on a salvsync request */
-#define VOL_STATE_SALVAGING      13      /* volume is being salvaged */
-#define VOL_STATE_ERROR          14      /* volume is in an error state */
-#define VOL_STATE_FREED          15      /* debugging aid */
+typedef enum {
+    VOL_STATE_UNATTACHED        = 0,    /**< volume is unattached */
+    VOL_STATE_PREATTACHED       = 1,    /**< volume has been pre-attached */
+    VOL_STATE_ATTACHING         = 2,    /**< volume is transitioning to fully attached */
+    VOL_STATE_ATTACHED          = 3,    /**< volume has been fully attached */
+    VOL_STATE_UPDATING          = 4,    /**< volume is updating on-disk structures */
+    VOL_STATE_GET_BITMAP        = 5,    /**< volume is getting bitmap entries */
+    VOL_STATE_HDR_LOADING       = 6,    /**< volume is loading disk header */
+    VOL_STATE_HDR_ATTACHING     = 7,    /**< volume is getting a header from the LRU */
+    VOL_STATE_SHUTTING_DOWN     = 8,    /**< volume is shutting down */
+    VOL_STATE_GOING_OFFLINE     = 9,    /**< volume is going offline */
+    VOL_STATE_OFFLINING         = 10,   /**< volume is transitioning to offline */
+    VOL_STATE_DETACHING         = 11,   /**< volume is transitioning to detached */
+    VOL_STATE_SALVSYNC_REQ      = 12,   /**< volume is blocked on a salvsync request */
+    VOL_STATE_SALVAGING         = 13,   /**< volume is being salvaged */
+    VOL_STATE_ERROR             = 14,   /**< volume is in an error state */
+    VOL_STATE_VNODE_ALLOC       = 15,   /**< volume is busy allocating a new vnode */
+    VOL_STATE_VNODE_GET         = 16,   /**< volume is busy getting vnode disk data */
+    VOL_STATE_VNODE_CLOSE       = 17,   /**< volume is busy closing vnodes */
+    VOL_STATE_VNODE_RELEASE     = 18,   /**< volume is busy releasing vnodes */
+    VOL_STATE_VLRU_ADD          = 19,   /**< volume is busy being added to a VLRU queue */
+    /* please add new states directly above this line */
+    VOL_STATE_FREED             = 20,   /**< debugging aid */
+    VOL_STATE_COUNT             = 21,   /**< total number of valid states */
+} VolState;
 
-#define VOL_STATE_COUNT          16      /* total number of valid states */
-
-/* V_attachFlags bits */
-#define VOL_HDR_ATTACHED   0x1     /* volume header is attached to Volume struct */
-#define VOL_HDR_LOADED     0x2     /* volume header contents are valid */
-#define VOL_HDR_IN_LRU     0x4     /* volume header is in LRU */
-#define VOL_IN_HASH        0x8     /* volume is in hash table */
-#define VOL_ON_VBYP_LIST   0x10    /* volume is on VByP list */
-#define VOL_IS_BUSY        0x20    /* volume is not to be free()d */
-#define VOL_ON_VLRU        0x40    /* volume is on the VLRU */
-#define VOL_HDR_DONTSALV   0x80    /* volume header DONTSALVAGE flag is set */
+/**
+ * V_attachFlags bits.
+ */
+enum VolFlags {
+    VOL_HDR_ATTACHED      = 0x1,     /**< volume header is attached to Volume struct */
+    VOL_HDR_LOADED        = 0x2,     /**< volume header contents are valid */
+    VOL_HDR_IN_LRU        = 0x4,     /**< volume header is in LRU */
+    VOL_IN_HASH           = 0x8,     /**< volume is in hash table */
+    VOL_ON_VBYP_LIST      = 0x10,    /**< volume is on VByP list */
+    VOL_IS_BUSY           = 0x20,    /**< volume is not to be free()d */
+    VOL_ON_VLRU           = 0x40,    /**< volume is on the VLRU */
+    VOL_HDR_DONTSALV      = 0x80,    /**< volume header DONTSALVAGE flag is set */
+};
 
 /* VPrintExtendedCacheStats flags */
-#define VOL_STATS_PER_CHAIN   0x1  /* compute simple per-chain stats */
-#define VOL_STATS_PER_CHAIN2  0x2  /* compute per-chain stats that require scanning
-				    * every element of the chain */
+#define VOL_STATS_PER_CHAIN   0x1  /**< compute simple per-chain stats */
+#define VOL_STATS_PER_CHAIN2  0x2  /**< compute per-chain stats that require scanning
+				    *   every element of the chain */
 
 /* VLRU_SetOptions options */
 #define VLRU_SET_THRESH       1
@@ -125,20 +204,35 @@ struct versionStamp {		/* Version stamp for critical volume files */
 #define VLRU_SET_MAX          3
 #define VLRU_SET_ENABLED      4
 
-/* valid VLRU queue names */
-#define VLRU_QUEUE_NEW 0            /* LRU queue for new volumes */
-#define VLRU_QUEUE_MID 1            /* survivor generation */
-#define VLRU_QUEUE_OLD 2            /* old generation */
-#define VLRU_QUEUE_CANDIDATE 3      /* soft detach candidate pool */
-#define VLRU_QUEUE_HELD 4           /* volumes which are not allowed
-				     * to be soft detached */
-#define VLRU_QUEUE_INVALID 5        /* invalid queue id */
+/**
+ * VLRU queue names.
+ */
+typedef enum {
+    VLRU_QUEUE_NEW        = 0,  /**< LRU queue for new volumes */
+    VLRU_QUEUE_MID        = 1,  /**< survivor generation */
+    VLRU_QUEUE_OLD        = 2,  /**< old generation */
+    VLRU_QUEUE_CANDIDATE  = 3,  /**< soft detach candidate pool */
+    VLRU_QUEUE_HELD       = 4,  /*   volumes which are not allowed
+				 *   to be soft detached */
+    VLRU_QUEUE_INVALID    = 5,  /**< invalid queue id */
+} VLRUQueueName;
 
 /* default scanner timing parameters */
 #define VLRU_DEFAULT_OFFLINE_THRESH (60*60*2) /* 2 hours */
 #define VLRU_DEFAULT_OFFLINE_INTERVAL (60*2) /* 2 minutes */
 #define VLRU_DEFAULT_OFFLINE_MAX 8 /* 8 volumes */
 
+
+/**
+ * DAFS thread-specific options structure
+ */
+typedef struct VThreadOptions {
+     int disallow_salvsync;     /**< whether or not salvsync calls are allowed
+				 *   on this thread (deadlock prevention). */
+} VThreadOptions_t;
+extern pthread_key_t VThread_key;
+extern VThreadOptions_t VThread_defaults;
+
 #endif /* AFS_DEMAND_ATTACH_FS */
 
 
@@ -364,7 +458,9 @@ typedef struct VolumeDiskData {
 /* Memory resident volume information */
 /**************************************/
 
-/* global volume package stats */
+/** 
+ * global volume package stats.
+ */
 typedef struct VolPkgStats {
 #ifdef AFS_DEMAND_ATTACH_FS
     /*
@@ -373,41 +469,37 @@ typedef struct VolPkgStats {
      */
 
     /* levels */
-    afs_uint32 state_levels[VOL_STATE_COUNT];
+    afs_uint32 state_levels[VOL_STATE_COUNT]; /**< volume state transition counters */
 
     /* counters */
-    afs_uint64 hash_looks;           /* number of hash chain element traversals */
-    afs_uint64 hash_reorders;        /* number of hash chain reorders */
-    afs_uint64 salvages;             /* online salvages since fileserver start */
-    afs_uint64 vol_ops;              /* volume operations since fileserver start */
+    afs_uint64 hash_looks;           /**< number of hash chain element traversals */
+    afs_uint64 hash_reorders;        /**< number of hash chain reorders */
+    afs_uint64 salvages;             /**< online salvages since fileserver start */
+    afs_uint64 vol_ops;              /**< volume operations since fileserver start */
 #endif /* AFS_DEMAND_ATTACH_FS */
 
-    afs_uint64 hdr_loads;            /* header loads from disk */
-    afs_uint64 hdr_gets;             /* header pulls out of LRU */
-    afs_uint64 attaches;             /* volume attaches since fileserver start */
-    afs_uint64 soft_detaches;        /* soft detach ops since fileserver start */
+    afs_uint64 hdr_loads;            /**< header loads from disk */
+    afs_uint64 hdr_gets;             /**< header pulls out of LRU */
+    afs_uint64 attaches;             /**< volume attaches since fileserver start */
+    afs_uint64 soft_detaches;        /**< soft detach ops since fileserver start */
 
     /* configuration parameters */
-    afs_uint32 hdr_cache_size;       /* size of volume header cache */
+    afs_uint32 hdr_cache_size;       /**< size of volume header cache */
 } VolPkgStats;
 extern VolPkgStats VStats;
 
 /*
  * volume header cache supporting structures
  */
-#ifdef AFS_DEMAND_ATTACH_FS
 struct volume_hdr_LRU_stats {
     afs_uint32 free;
     afs_uint32 used;
     afs_uint32 attached;
 };
-#endif
 
 struct volume_hdr_LRU_t {
     struct rx_queue lru;
-#ifdef AFS_DEMAND_ATTACH_FS
     struct volume_hdr_LRU_stats stats;
-#endif
 };
 extern struct volume_hdr_LRU_t volume_hdr_LRU;
 
@@ -452,51 +544,52 @@ struct VolumeHashChainStats {
 
 
 #ifdef AFS_DEMAND_ATTACH_FS
-/* demand attach fs
- * extended per-volume statistics 
+/**
+ * DAFS extended per-volume statistics.
  *
- * please note that this structure lives across the entire
- * lifetime of the fileserver process
+ * @note this data lives across the entire
+ *       lifetime of the fileserver process
  */
 typedef struct VolumeStats {
     /* counters */
-    afs_uint64 hash_lookups;         /* hash table lookups */
-    afs_uint64 hash_short_circuits;  /* short circuited hash lookups (due to cacheCheck) */
-    afs_uint64 hdr_loads;            /* header loads from disk */
-    afs_uint64 hdr_gets;             /* header pulls out of LRU */
-    afs_uint16 attaches;             /* attaches of this volume since fileserver start */
-    afs_uint16 soft_detaches;        /* soft detaches of this volume */
-    afs_uint16 salvages;             /* online salvages since fileserver start */
-    afs_uint16 vol_ops;              /* volume operations since fileserver start */
+    afs_uint64 hash_lookups;         /**< hash table lookups */
+    afs_uint64 hash_short_circuits;  /**< short circuited hash lookups (due to cacheCheck) */
+    afs_uint64 hdr_loads;            /**< header loads from disk */
+    afs_uint64 hdr_gets;             /**< header pulls out of LRU */
+    afs_uint16 attaches;             /**< attaches of this volume since fileserver start */
+    afs_uint16 soft_detaches;        /**< soft detaches of this volume */
+    afs_uint16 salvages;             /**< online salvages since fileserver start */
+    afs_uint16 vol_ops;              /**< volume operations since fileserver start */
 
     /* timestamps */
-    afs_uint32 last_attach;      /* unix timestamp of last VAttach */
-    afs_uint32 last_get;         /* unix timestamp of last VGet/VHold */
-    afs_uint32 last_promote;     /* unix timestamp of last VLRU promote/demote */
-    afs_uint32 last_hdr_get;     /* unix timestamp of last GetVolumeHeader() */
-    afs_uint32 last_salvage;     /* unix timestamp of last initiation of an online salvage */
-    afs_uint32 last_salvage_req; /* unix timestamp of last SALVSYNC request */
-    afs_uint32 last_vol_op;      /* unix timestamp of last volume operation */
+    afs_uint32 last_attach;      /**< unix timestamp of last VAttach */
+    afs_uint32 last_get;         /**< unix timestamp of last VGet/VHold */
+    afs_uint32 last_promote;     /**< unix timestamp of last VLRU promote/demote */
+    afs_uint32 last_hdr_get;     /**< unix timestamp of last GetVolumeHeader() */
+    afs_uint32 last_hdr_load;    /**< unix timestamp of last LoadVolumeHeader() */
+    afs_uint32 last_salvage;     /**< unix timestamp of last initiation of an online salvage */
+    afs_uint32 last_salvage_req; /**< unix timestamp of last SALVSYNC request */
+    afs_uint32 last_vol_op;      /**< unix timestamp of last volume operation */
 } VolumeStats;
 
-/* demand attach fs
- * online salvager state */
+/**
+ * DAFS online salvager state.
+ */
 typedef struct VolumeOnlineSalvage {
-    afs_uint32 prio;            /* number of VGetVolume's since salvage requested */
-    int reason;                 /* reason for requesting online salvage */
-    byte requested;             /* flag specifying that salvage should be scheduled */
-    byte scheduled;             /* flag specifying whether online salvage scheduled */
-    byte reserved[2];           /* padding */
+    afs_uint32 prio;            /**< number of VGetVolume's since salvage requested */
+    int reason;                 /**< reason for requesting online salvage */
+    byte requested;             /**< flag specifying that salvage should be scheduled */
+    byte scheduled;             /**< flag specifying whether online salvage scheduled */
+    byte reserved[2];           /**< padding */
 } VolumeOnlineSalvage;
 
-/* demand attach fs
- * volume LRU state */
+/**
+ * DAFS Volume LRU state.
+ */
 typedef struct VolumeVLRUState {
-    struct rx_queue lru;        /* VLRU queue pointers */
-    int idx;                    /* VLRU generation index */
+    struct rx_queue lru;        /**< VLRU queue for this generation */
+    VLRUQueueName idx;          /**< VLRU generation index */
 } VolumeVLRUState;
-
-typedef afs_uint16 VolState;    /* attachment state type */
 #endif /* AFS_DEMAND_ATTACH_FS */
 
 typedef struct Volume {
@@ -504,7 +597,7 @@ typedef struct Volume {
     VolumeId hashid;		/* Volume number -- for hash table lookup */
     struct volHeader *header;	/* Cached disk data */
     Device device;		/* Unix device for the volume */
-    struct DiskPartition
+    struct DiskPartition64
      *partition;		/* Information about the Unix partition */
     struct vnodeIndex {
 	IHandle_t *handle;	/* Unix inode holding this index */
@@ -538,12 +631,14 @@ typedef struct Volume {
     byte specialStatus;		/* An error code to return on VGetVolume: the
 				 * volume is unavailable for the reason quoted,
 				 * currently VBUSY or VMOVED */
+    afs_uint32 checkoutMode;    /* for volume utilities, mode number for current checkout */
     afs_uint32 updateTime;	/* Time that this volume was put on the updated
 				 * volume list--the list of volumes that will be
 				 * salvaged should the file server crash */
+    struct rx_queue vnode_list; /**< linked list of cached vnodes for this volume */
 #ifdef AFS_DEMAND_ATTACH_FS
     VolState attach_state;      /* what stage of attachment has been completed */
-    afs_uint16 attach_flags;    /* flags related to attachment state */
+    afs_uint32 attach_flags;    /* flags related to attachment state */
     pthread_cond_t attach_cv;   /* state change condition variable */
     short nWaiters;             /* volume package internal ref count */
     int chainCacheCheck;        /* Volume hash chain cache check */
@@ -571,6 +666,7 @@ struct volHeader {
 #define V_vnodeIndex(vp)	((vp)->vnodeIndex)
 #define V_nextVnodeUnique(vp)	((vp)->nextVnodeUnique)
 #define V_linkHandle(vp)	((vp)->linkHandle)
+#define V_checkoutMode(vp)      ((vp)->checkoutMode)
 #ifdef AFS_DEMAND_ATTACH_FS
 #define V_attachState(vp)       ((vp)->attach_state)
 #define V_attachFlags(vp)       ((vp)->attach_flags)
@@ -680,8 +776,8 @@ extern void VSetDiskUsage(void);
 extern void VPrintCacheStats(void);
 extern void VReleaseVnodeFiles_r(Volume * vp);
 extern void VCloseVnodeFiles_r(Volume * vp);
-extern struct DiskPartition *VGetPartition(char *name, int abortp);
-extern struct DiskPartition *VGetPartition_r(char *name, int abortp);
+extern struct DiskPartition64 *VGetPartition(char *name, int abortp);
+extern struct DiskPartition64 *VGetPartition_r(char *name, int abortp);
 extern int VInitVolumePackage(ProgramType pt, afs_uint32 nLargeVnodes,
 			      afs_uint32 nSmallVnodes, int connect, afs_uint32 volcache);
 extern void DiskToVolumeHeader(VolumeHeader_t * h, VolumeDiskHeader_t * dh);
@@ -689,16 +785,17 @@ extern void VolumeHeaderToDisk(VolumeDiskHeader_t * dh, VolumeHeader_t * h);
 extern void VTakeOffline_r(register Volume * vp);
 extern void VTakeOffline(register Volume * vp);
 extern Volume * VLookupVolume_r(Error * ec, VolId volumeId, Volume * hint);
-
+extern void VGetVolumePath(Error * ec, VolId volumeId, char **partitionp,
+			   char **namep);
 #ifdef AFS_DEMAND_ATTACH_FS
-extern Volume *VPreAttachVolumeByName(Error * ec, char *partition, char *name, 
-				       int mode);
-extern Volume *VPreAttachVolumeByName_r(Error * ec, char *partition, char *name,
-				     int mode);
-extern Volume *VPreAttachVolumeById_r(Error * ec, struct DiskPartition * partp, 
-				      Volume * vp, int volume_id);
+extern Volume *VPreAttachVolumeByName(Error * ec, char *partition, char *name);
+extern Volume *VPreAttachVolumeByName_r(Error * ec, char *partition, char *name);
+extern Volume *VPreAttachVolumeById_r(Error * ec, char * partition, 
+				      VolId volumeId);
+extern Volume *VPreAttachVolumeByVp_r(Error * ec, struct DiskPartition64 * partp, 
+				      Volume * vp, VolId volume_id);
 extern Volume *VGetVolumeByVp_r(Error * ec, Volume * vp);
-extern int VShutdownByPartition_r(struct DiskPartition * dp);
+extern int VShutdownByPartition_r(struct DiskPartition64 * dp);
 extern int VShutdownVolume_r(Volume * vp);
 extern int VConnectSALV(void);
 extern int VConnectSALV_r(void);
@@ -708,12 +805,12 @@ extern int VDisconnectSALV(void);
 extern int VDisconnectSALV_r(void);
 extern void VPrintExtendedCacheStats(int flags);
 extern void VPrintExtendedCacheStats_r(int flags);
-extern VolState VChangeState_r(Volume * vp, VolState new_state);
 extern void VLRU_SetOptions(int option, afs_uint32 val);
 extern int VSetVolHashSize(int logsize);
-extern int VRequestSalvage_r(Volume * vp, int reason, int flags);
+extern int VRequestSalvage_r(Error * ec, Volume * vp, int reason, int flags);
 extern int VRegisterVolOp_r(Volume * vp, FSSYNC_VolOp_info * vopinfo);
-extern int VDeregisterVolOp_r(Volume * vp, FSSYNC_VolOp_info * vopinfo);
+extern int VDeregisterVolOp_r(Volume * vp);
+extern void VCancelReservation_r(Volume * vp);
 #endif /* AFS_DEMAND_ATTACH_FS */
 extern int VVolOpLeaveOnline_r(Volume * vp, FSSYNC_VolOp_info * vopinfo);
 extern int VVolOpSetVBusy_r(Volume * vp, FSSYNC_VolOp_info * vopinfo);
